/////////////////////////////////////////////////////////////
// cmdasuser.cpp
//
// Copyright 1999, Keith Brown
// http://www.develop.com/kbrown
//
// Command line utility that executes a command shell as
// the specified user. Loads the user's profile (creating one
// if necessary), and creates a correct environment block
// for the new process. Demonstrates manipulation of winsta
// and desktop DACLs, granting access to logon session SIDs.
//
// To execute this application, you must be an administrator.
// The principal you specify as a command line argument
// must have the "Log on Locally" privilege on the machine
// (unless you're running on a domain conroller, everyone has
// this privilege by default, so this shouldn't be an issue).
//
// Tested on Windows NT 4.0 SP3, SP4, and Windows 2000 Beta 3.
//
// Usage: cmdasuser authority\principal password
//        cmdasuser localsystem
//
/////////////////////////////////////////////////////////////
#define _WIN32_WINNT 0x400
//#define UNICODE
#include <windows.h>
#include <stdio.h>
#include <aclapi.h>
//#include <ACCCTRL.H>
// #include "cmdTempService.h"

// userenv.h can be found in the Windows 2000 beta SDK,
// and you'll need it to build this project...
#include "userenv.h"

#ifdef __cplusplus
extern "C" {
#endif

// for some reason, winuser.h doesn't define these useful aliases
#ifndef WINSTA_ALL_ACCESS
#define WINSTA_ALL_ACCESS 0x0000037F
#endif

#ifndef DESKTOP_ALL_ACCESS
#define DESKTOP_ALL_ACCESS 0x00001FF
#endif

//HANDLE g_hProcessHeap = GetProcessHeap();
char *err;
char pszFcn[64];

void Err( const char* pszFcn1, DWORD nErr = GetLastError() )
{
    char szErrMsg[256];
    char szMsg[512];
    if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, nErr, 0, szErrMsg, sizeof szErrMsg / sizeof *szErrMsg, 0 ) )
         wsprintf( szMsg, "%s failed: %s", pszFcn, szErrMsg );
    else wsprintf( szMsg, "%s failed: 0x%08X", pszFcn, nErr );
    MessageBox(NULL, szMsg, "as user", 0);
    exit(nErr);
    lstrcpy(pszFcn, pszFcn1);
    
    throw (int)GetLastError();
}

bool instr( const char* psz, char c )
{
    while ( *psz )
        if ( *psz++ == c )
            return true;
    return false;
}

ACL* GetUserObjectDacl( HANDLE h, DWORD cbExtra )
{
    // At least one person complained about GetSecurityInfo failing with
    // E_INVALIDARG, so I changed the code to use the older API
    // and this seemed to fix the problem.
    // But the GetSecurityInfo version was soooo much cleaner :-)
    DWORD cb = 0;
    SECURITY_INFORMATION si = DACL_SECURITY_INFORMATION;
    GetUserObjectSecurity( h, &si, 0, cb, &cb );
    if ( ERROR_INSUFFICIENT_BUFFER != GetLastError() )
        Err( "GetUserObjectSecurity" );

    HANDLE hProcessHeap = GetProcessHeap();

    void* psd = HeapAlloc( hProcessHeap, 0, cb );
    if ( !psd )
        Err( "HeapAlloc", E_OUTOFMEMORY );

    si = DACL_SECURITY_INFORMATION;
    if ( !GetUserObjectSecurity( h, &si, psd, cb, &cb ) )
        Err( "GetUserObjectSecurity" );

    BOOL bPresent, bDefaulted;
    ACL* pdaclOld;
    if ( !GetSecurityDescriptorDacl( psd, &bPresent, &pdaclOld, &bDefaulted ) )
        Err( "GetSecurityDescriptorDac" );

    // get the size of the existing DACL
    ACL_SIZE_INFORMATION info;
    if ( !GetAclInformation( pdaclOld, &info, sizeof info, AclSizeInformation ) )
        Err( "GetAclInformation" );

    // allocate enough memory for the existing DACL
    // plus whatever extra space the caller requires
    cb = info.AclBytesInUse + cbExtra;
    ACL* pdaclNew = reinterpret_cast<ACL*>( HeapAlloc( hProcessHeap, 0, cb ) );
    if ( !pdaclNew )
        Err( "HeapAlloc" );

    if ( !InitializeAcl( pdaclNew, cb, ACL_REVISION ) )
        Err( "InitializeAc" );
    
    // copy over all the old aces to the new DACL
    for ( DWORD i = 0; i < info.AceCount; ++i )
    {
        ACE_HEADER* pace = 0;
        if ( !GetAce( pdaclOld, i, reinterpret_cast<void**>(&pace) ) )
            Err( "GetAce" );

        if ( !AddAce( pdaclNew, ACL_REVISION, MAXDWORD, pace, pace->AceSize ) )
            Err( "AddAce" );
    }
    HeapFree( hProcessHeap, 0, psd );
    return pdaclNew;
}

DWORD CalcAceSizeFromSid( void* psid )
{
    return sizeof( ACCESS_ALLOWED_ACE ) - sizeof( DWORD ) + GetLengthSid( psid );
}

// this is such a simple function, it's bizarre that we had to wait 'till W2K to get it
BOOL AddAccessAllowedAceEx4( PACL pAcl, DWORD dwAceRevision, DWORD AceFlags, DWORD AccessMask, PSID pSid )
{
    if ( !AddAccessAllowedAce( pAcl, dwAceRevision, AccessMask, pSid ) )
        return FALSE;
    ACL_SIZE_INFORMATION info;
    if ( !GetAclInformation( pAcl, &info, sizeof info, AclSizeInformation ) )
        return FALSE;
    ACE_HEADER* pace = 0;
    if ( !GetAce( pAcl, info.AceCount - 1, reinterpret_cast<void**>(&pace) ) )
        return FALSE;
    pace->AceFlags = static_cast<BYTE>(AceFlags);
    return TRUE;
}
//BOOL AddAccessAllowedAceEx( PACL pAcl, DWORD dwAceRevision, DWORD AceFlags, DWORD AccessMask, PSID pSid );

void DeleteMatchingAces( ACL* pdacl, void* psid )
{
    ACL_SIZE_INFORMATION info;
    if ( !GetAclInformation( pdacl, &info, sizeof info, AclSizeInformation ) )
        Err( "GetAclInformation" );

    // it's a bit easier to delete aces while iterating backwards
    // so that the iterator doesn't get honked up
    DWORD i = info.AceCount;
    while ( i-- )
    {
        ACCESS_ALLOWED_ACE* pAce = 0;
        if ( !GetAce( pdacl, i, reinterpret_cast<void**>(&pAce) ) )
            Err( "GetAce" );
        if ( EqualSid( psid, &pAce->SidStart ) )
            DeleteAce( pdacl, i );
    }
}

void GrantSessionSIDAccessToWinstationAndDesktop( HANDLE htok, bool bGrant )
{
    BYTE tgs[4096];
    DWORD cbtgs = sizeof tgs;
    HANDLE hProcessHeap = GetProcessHeap();
    if ( !GetTokenInformation( htok, TokenGroups, tgs, cbtgs, &cbtgs ) )
        Err( "GetTokenInformation" );
    const TOKEN_GROUPS* ptgs = reinterpret_cast<TOKEN_GROUPS*>(tgs);
    const SID_AND_ATTRIBUTES* it = ptgs->Groups;
    const SID_AND_ATTRIBUTES* end = it + ptgs->GroupCount;
    while ( end != it )
    {
        if ( it->Attributes & SE_GROUP_LOGON_ID )
            break;
        ++it;
    }
    if ( end == it )
        Err( "UNEXPECTED: No Logon SID in TokenGroups" );

    void* psidLogonSession = it->Sid;

    HWINSTA hws = GetProcessWindowStation();
    if ( !hws )
        Err( "GetProcessWindowStation" );

    {
        ACL* pdacl = 0;
        if ( bGrant )
        {
            // get the existing DACL, with enough extra space for adding a couple more aces
            const DWORD cbExtra = 2 * CalcAceSizeFromSid( psidLogonSession );
            pdacl = GetUserObjectDacl( hws, cbExtra );

            // grant the logon session all access to winsta0
            if ( !AddAccessAllowedAce( pdacl, ACL_REVISION,
                                        WINSTA_ALL_ACCESS | STANDARD_RIGHTS_REQUIRED,
                                        psidLogonSession ) )
                Err( "AddAccessAllowedAce" );

            // grant the logon session all access to any new desktops created in winsta0
            // by adding an inherit-only ace
            int ver=0xFF & GetVersion();
            int res=0;
            if ( !AddAccessAllowedAceEx4( pdacl, ACL_REVISION,
                                        SUB_CONTAINERS_AND_OBJECTS_INHERIT | INHERIT_ONLY,
                                        DESKTOP_ALL_ACCESS | STANDARD_RIGHTS_REQUIRED,
                                        psidLogonSession ))
                Err( "AddAccessAllowedAceEx" );
        }
        else
        {
            pdacl = GetUserObjectDacl( hws, 0 );
            DeleteMatchingAces( pdacl, psidLogonSession );
        }

        // apply the changes to winsta0
        // unlike GetSecurityInfo, SetSecurityInfo has no troubles on SP3 :-)
        DWORD err = SetSecurityInfo( hws, SE_WINDOW_OBJECT, DACL_SECURITY_INFORMATION, 0, 0, pdacl, 0 );
        if ( err )
            Err( "SetSecurityInfo", err );

        HeapFree( hProcessHeap, 0, pdacl );
    }

    CloseWindowStation( hws );

    HDESK hd = GetThreadDesktop( GetCurrentThreadId() );
    if ( !hd )
        Err( "GetThreadDesktop" );

    {
        ACL* pdacl = 0;
        if ( bGrant )
        {
            // get the existing DACL, with enough extra space for adding a couple more aces
            const DWORD cbExtra = CalcAceSizeFromSid( psidLogonSession );
            pdacl = GetUserObjectDacl( hd, cbExtra );

            // grant the logon session all access to the default desktop
            if ( !AddAccessAllowedAce( pdacl, ACL_REVISION,
                                        DESKTOP_ALL_ACCESS | STANDARD_RIGHTS_REQUIRED,
                                        psidLogonSession ) )
                Err( "AddAccessAllowedAce" );

        }
        else
        {
            pdacl = GetUserObjectDacl( hd, 0 );
            DeleteMatchingAces( pdacl, psidLogonSession );
        }

        // apply the changes to the default desktop
        // unlike GetSecurityInfo, SetSecurityInfo has no troubles on SP3 :-)
        DWORD err = SetSecurityInfo( hd, SE_WINDOW_OBJECT, DACL_SECURITY_INFORMATION, 0, 0, pdacl, 0 );
        if ( err )
            Err( "SetSecurityInfo", err );

        HeapFree( hProcessHeap, 0, pdacl );
    }

    CloseDesktop( hd );
}

typedef BOOL (WINAPI * CREATE_ENVIRONMENT_BLOCK) ( void**, HANDLE, BOOL );
typedef BOOL (WINAPI * DESTROY_ENVIRONMENT_BLOCK) ( void* );
typedef BOOL (WINAPI * LOAD_USER_PROFILE) ( HANDLE, PROFILEINFO* );
typedef BOOL (WINAPI * UNLOAD_USER_PROFILE) ( HANDLE, HANDLE );
CREATE_ENVIRONMENT_BLOCK    m_createEnvironmentBlock;
DESTROY_ENVIRONMENT_BLOCK   m_destroyEnvironmentBlock;
LOAD_USER_PROFILE           m_loadUserProfile;
UNLOAD_USER_PROFILE         m_unloadUserProfile;


HMODULE m_hModule;

 void _getAddr( const char* psz, FARPROC* pProc ){
     *pProc = (FARPROC)GetProcAddress( m_hModule, psz ) ;
     if ( !*pProc )
         Err( "GetProcAddress" );
 }

void UserEnvInterface(){
    m_hModule = LoadLibrary( "userenv.dll" );
    if ( !m_hModule )
        Err( "LoadLibrary( userenv.dll )" );

    _getAddr( "CreateEnvironmentBlock",  (FARPROC *)&m_createEnvironmentBlock );
    _getAddr( "DestroyEnvironmentBlock", (FARPROC *)&m_destroyEnvironmentBlock );
    _getAddr( "LoadUserProfileW",        (FARPROC *)&m_loadUserProfile );
    _getAddr( "UnloadUserProfile",       (FARPROC *)&m_unloadUserProfile );
}

// STEP 3: we're now inside the "monitor" process which will block
//         until the target process dies, allowing us to clean up
//         the profile we loaded, as well as the winsta/desktop dacls.
LRESULT CmdAsUser( char* pszPrincipal, const char* pszPassword, char* pszAuthority, char* szcmd)
{

    HANDLE htok = 0;
    if ( !LogonUser(    const_cast<char*>( pszPrincipal ),
                        const_cast<char*>( pszAuthority ),
                        const_cast<char*>( pszPassword ),
                        LOGON32_LOGON_INTERACTIVE,
                        LOGON32_PROVIDER_DEFAULT,
                        &htok ) )
        Err( "LogonUser" );

    // adjust the interactive winsta/desktop DACLs
    GrantSessionSIDAccessToWinstationAndDesktop( htok, true );

    // load userenv.dll and map the entrypoints we need
    UserEnvInterface();

    // for some very strange reason, LoadUserProfile needs
    // not only the *token* for the user, but also the user's name...
    // I sure wish there was an API for getting the user name without
    // having to impersonate (LookupAccountSid is expensive for domain accounts)
    char szUserName[1024]; DWORD cchUserName = sizeof szUserName / sizeof *szUserName;
    if ( !ImpersonateLoggedOnUser( htok ) )
        Err( "ImpersonateLoggedOnUser" );
    if ( !GetUserName( szUserName, &cchUserName ) ) 
        Err( "GetUserName" );
    RevertToSelf();

    // load the user profile
    PROFILEINFO profinfo = { sizeof profinfo, 0, szUserName };
    if ( !m_loadUserProfile( htok, &profinfo ) )
        Err( "LoadUserProfile" );

    // set up an environment block
    void* pEnvBlock = 0;
    if ( !m_createEnvironmentBlock( &pEnvBlock, htok, FALSE ) )
        Err( "CreateEnvironmentBlock" );

    // here's where we start the target process (cmd.exe)
    // note that we give the command shell a custom title for clarity
//    char szcmd[256];
//    wsprintf( szcmd, "cmd.exe /K title %s\\%s", pszAuthority, pszPrincipal );
    STARTUPINFO si = { sizeof si };
    PROCESS_INFORMATION pi;
    if ( !CreateProcessAsUser( htok, 0, szcmd, 0, 0, FALSE, CREATE_NEW_CONSOLE | CREATE_UNICODE_ENVIRONMENT,
                               pEnvBlock, 0, &si, &pi ) )
        Err( "CreateProcessAsUser" );

    WaitForSingleObject( pi.hProcess, INFINITE );
    LRESULT code;
    if(!GetExitCodeProcess(pi.hProcess, (unsigned long *)&code))
        code=GetLastError();
    CloseHandle( pi.hThread );
    CloseHandle( pi.hProcess );

    // free the environment block
    m_destroyEnvironmentBlock( pEnvBlock );

    // unload the user profile
    m_unloadUserProfile( htok, profinfo.hProfile );

    // clean up the interactive winsta/desktop DACLs
    GrantSessionSIDAccessToWinstationAndDesktop( htok, false );

    CloseHandle( htok );
    return code;
}


LRESULT CALLBACK AsUser(char *user, char *pass, char *domain, char *cmd){
    try{
        return CmdAsUser(user, pass, domain, cmd);
    }
    catch(int nErr){
       char szErrMsg[256];
       char szMsg[512];
       if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, nErr, 0, szErrMsg, sizeof szErrMsg / sizeof *szErrMsg, 0 ) )
            wsprintf( szMsg, "%s failed: %s", pszFcn, szErrMsg );
       else wsprintf( szMsg, "%s failed: 0x%08X", pszFcn, nErr );
       MessageBox(NULL, szMsg, "as user", 0);
       return nErr;
    }
}

#ifdef __cplusplus
}
#endif
