{ ====================================================================
(C) Copyright 1999 FORTH, Inc.   www.forth.com

PACKAGES -- an alternative to vocabularies
==================================================================== }

{ --------------------------------------------------------------------
Encapsulation is the process of containing a set of entities such that
the members are only visible thru a user-defined window.  Object
oriented programming is one kind of encapsulation. Another is when a
word or routine requires supporting words for its definition, but which
have no interest to the "outside" world.

-PACKAGE enforces a valid PACKAGE frame on the data stack. The
   frame is considered valid iff: the tag is the xt of wordlist,
   and the value returned by GET-CURRENT is either the wid of
   the package or its parent.

?PACKAGE checks the frame, and aborts if bad.

OPEN-PACKAGE creates a frame given a WID.

PACKAGE creates or re-uses a package, leaving a valid frame.

END-PACKAGE restores CONTEXT and CURRENT to their prior states.

PUBLIC changes CURRENT to the parent and
PRIVATE changes CURRENT to the package.
-------------------------------------------------------------------- }

?( PACKAGES -- an alternative to vocabularies)

: -ORDER ( wid -- )
    >R GET-ORDER 0
    BEGIN 2DUP > WHILE
        ( x x ... # i )
        DUP 2+ PICK R@ =
        IF DUP 2+ ROLL DROP 
           SWAP 1- SWAP
        ELSE
           1+
        THEN
    REPEAT
    DROP
    ?DUP 0= IF FORTH-WORDLIST 1 THEN
    SET-ORDER
    RDROP
;
: +ORDER ( wid -- )
    DUP -ORDER >R GET-ORDER 1+ R> SWAP SET-ORDER ;


\ THROW#
(   S" Package balance changed") 100000 CONSTANT IOR_PACKAGEBALANCE
(   S" Not a package")           100001 CONSTANT IOR_PACKAGENOT
\ TO THROW#

: -PACKAGE ( parent package tag -- flag )   ['] WORDLIST =
   SWAP GET-CURRENT = ROT GET-CURRENT = OR AND 0= ;

: ?PACKAGE ( parent package tag -- parent package tag )
   3DUP -PACKAGE IOR_PACKAGEBALANCE ?THROW ;

: OPEN-PACKAGE ( wid -- parent package tag )
   GET-CURRENT  SWAP DUP SET-CURRENT
   DUP +ORDER  ['] WORDLIST ;


: PACKAGE ( -- parent package tag )
   >IN @  BL WORD FIND IF ( exists)  NIP
      DUP >BODY CELL+ @ ['] WORDLIST <> IOR_PACKAGENOT ?THROW
      >BODY @ OPEN-PACKAGE  EXIT
   THEN  DROP >IN !
   VOCABULARY
   ['] WORDLIST ,
   LATEST NAME> >BODY @ OPEN-PACKAGE
   DOES> @ ;

: END-PACKAGE ( parent package tag -- )
   ?PACKAGE DROP -ORDER SET-CURRENT ;

: PUBLIC ( parent package tag -- parent package tag )
   ?PACKAGE  THIRD SET-CURRENT ;

: PRIVATE ( parent package tag -- parent package tag )
   ?PACKAGE OVER SET-CURRENT ;

{ --------------------------------------------------------------------
/FORTH sets the default search order
-------------------------------------------------------------------- }

?( ... /FORTH sets the default search order)

: /FORTH
   ONLY FORTH-WORDLIST +ORDER ALSO DEFINITIONS ;




