\ File ANS1V1.SEQ from the set of files FOSM1V1.ZIP Nov-96 containing:
\      README.TXT Summary of FOSM1V1.ZIP
\     INSTALL.TXT Instructions for loading and running FoSM
\    TUTOR1V1.SEQ Introduction and tutorial to FoSM v1.1
\      ANS1V1.SEQ FoSM v1.1 in ANS Forth (this file) including test cases
\      ANY1V1.SEQ FoSM needs a small set of words for manipulating Return
\                 Addresses.  The ones in this file are suitable for most
\                 Forths.  Check them and load the file before ANS1V1.SEQ.
\      FPC1V1.SEQ ANS Forth words written in F-PC for DOS.  If using F-PC,
\                 then load this file instead of ANY1V1.SEQ.
\    INNER1V1.TXT The inner workings of FoSM v1.1.
\     FORML96.TXT Published paper on FoSM and a portable word set for
\                 manipulating Return Addresses.
\      GEMFST.SEQ Gordon Charlton's FoSM for MPE's GEMForth/ST (Atari) with
\                 examples as published in EuroFORML '91 and '92 (25 pages)
\    GCEXAMPL.SEQ Examples from GEMFST.SEQ ported to ANS Forth.
\ 
\ Copyright: FoSM, A Forth String Matcher, was written by Gordon Charlton
\ in '91 and ported to ANS Forth by Chris Jakeman in '96.  Both are members of
\ FIG UK.  Permission is given for the code to be freely used provided its
\ origin is acknowledged.

\ Acknowledgements: Thanks to contributors to comp.lang.forth for advice on
\ porting to ANS.

\ : # ; IMMEDIATE                         \ # is just a marker and does nothing


\ CONTENTS
\       Changes Since Previous Version
\       Portability
\       Porting To Non-ANS Forths
\       FoSM Code
\       Optional FoSM Code
\       Test Cases
\       Appendix
\         - Changes From EuroFORML Version
\         - Suggestions For Next Version

\ CHANGES SINCE PREVIOUS VERSION
\ Changes since version 1.0 are marked with "#".  These include:
\     - re-working $=? to avoid  0 C@  which some OSes don't like
\     - adding >N, N@, N> to promote variable-free coding
\     - adding $Any and $None
\     - adding AnyOf$

\ PORTABILITY
\ FoSM is an ANS Forth Program with Environmental Dependencies:
\ - requiring NIP TUCK WITHIN ?DO \ from the Core Extensions
\    word set.
\ - requiring CS-ROLL from the Control Flow word set.
\ - using lower case for standard definition names.
\ - requiring the non-standard words below which manipulate Return
\   Addresses.
\   RA-sys below indicates an implementation-defined return address
\   which may be more than 1 cell wide:
\   - RAStash   ( -- ) Allow access to Return Stack by removing RA-sys
\                      if kept on Return Stack.  Only one RA-sys is
\                      stashed at any time.
\   - RAUnstash ( -- ) Undo RAStash.
\   - RAGet     ( -- RA-sys ) Move RA-sys to Data Stack.
\   - RACopy    ( -- RA-sys ) Copy RA-sys to Data Stack.
\   - RAPut     ( RA-sys -- ) Move RA-sys back from Data Stack.
\   - RADrop    ( RA-sys -- ) Drop RA-sys from the Data Stac

\ Porting was checked by testing on a system with : CHARS 2* ;


\ PORTING TO NON-ANS FORTHS
\ If you are using F-PC, then load file FPC1V0.SEQ, which will load some
\ ANS words and then this file.  It will also set FPC: (defined below) to
\ change the definition of << below.
\
\ To port the difficult words << .. >> and EVALUATE, see GEMFST.SEQ
\ above for an implementation which uses indirect-threading and an outer
\ interpreter with co-routines.


\ CODE
\ This comes in 3 parts:
\       1. Basic FoSM
\       2. Optional extensions for
\          a Anonymous patterns
\          b Combinations
\          c Permutations
\       3. Test words to validate FoSM

\ --------------------------------------------------------------------------

REQUIRE ROLL lib/ext/core-ext.f

HERE
CR .( Compiling essential FoSM ) CR

\ CR .( Loading RA words.)

VARIABLE RATEMP                 \ TO HOLD RETURN ADDRESSES TEMPORARILY.

                                \ GIVE ACCESS TO THE RETURN STACK BY STASHING
                                \ THE NEXT RETURN ADDRESS IN A TEMPORARY
                                \ VARIABLE.
: RASTASH  ( -- )
        POSTPONE R>  POSTPONE RATEMP  POSTPONE ! ; IMMEDIATE

: RAUNSTASH  ( -- )
        POSTPONE RATEMP  POSTPONE @  POSTPONE >R ; IMMEDIATE

                                \ GET AND COPY THE NEXT RA TO THE DATA STACK
                                \ OR PUT IT BACK FROM THE DATA STACK.
: RAGET  ( -- RA )
        POSTPONE R> ; IMMEDIATE

: RACOPY  ( -- RA )
        POSTPONE R@ ; IMMEDIATE

: RAPUT  ( -- RA )
        POSTPONE >R ; IMMEDIATE

: RADROP  ( -- )
        POSTPONE DROP ; IMMEDIATE

\ CR .( RA words loaded.  Now load ANS1V1.SEQ.)

: CS-ROLL
        1+ 2* DUP ROLL
        SWAP 1- ROLL ;

\ FoSM uses NOT in the sense of 0=.
: NOT ( x -- flag ) 0= ;

        \ The following 2 words allow conditional compilation for FPC if
        \ file FPC1V0.SEQ has been loaded to provide FPCDEFINED.
: DEFINED? ( ++ -- FLAG )                                               \ CMJ
        BL WORD FIND NIP ;

: FPC: ( -- )                                                           \ CMJ
        [ DEFINED? FPCDEFINED ] LITERAL NOT IF POSTPONE \ THEN ; IMMEDIATE


                                        \ SET TRUE TO COMPILE TEST CASES AND
                                        \ PRINT PROGRESS OF EACH MATCH.
                                        \ SET FALSE TO IGNORE FROM FD: TO THE
                                        \ END OF A LINE.
FALSE CONSTANT FOSMDEBUG                                                 \ CMJ
                                        \ FD: COMPILES IN EXTRA DEBUG CODE
                                        \ WHEN FOSMDEBUG IS SET TRUE.  THE
                                        \ DEBUG PRINTS THE PROGRESS OF THE
                                        \ MATCH.  FOR EXAMPLE:
                                        \     << PATTERN { B } A >>
                                        \     TEST P A# PASSED?
                                        \ PRINTS:
                                        \   << PATTERN { B } A >> - PATTERN
                                        \   "A"                   - TEXT
                                        \    <<
                                        \    {                    - OPTION
                                        \    B - BACKTRACK TO {   - MISMATCH
                                        \    }                    - OPTION END
                                        \    A => MATCH           - MATCH ENDS

: FD: ( -- )                                                            \ CMJ
        FOSMDEBUG NOT IF POSTPONE \ THEN ; IMMEDIATE

                                \ SKIP THE WORD WHICH CALLED THIS ONE BY
                                \ REMOVING ITS RETURN ADDRESS.
: SKIPCALLER ( -- )
        POSTPONE RAGET  POSTPONE RADROP ; IMMEDIATE

: CHARS/ ( #ADDRESSUNITS -- #CHARS )    \ RECIPROCAL OF CHARS
        1 CHARS  1 =
        IF ( IE "1 /" SO DO NOTHING )
        ELSE S" [ 1 CHARS ] LITERAL /" EVALUATE
        THEN ; IMMEDIATE


\ FOSM  --  UTILITIES  --  STACK  ARRAY --------------------------------------
                                        \ CREATES A NAMED STACK.
                                        \ 1ST CELL HOLDS &STACKTOP
                                        \ NOTE THAT PUSH AND POP DO NOT
                                        \ INCLUDE ANY ERROR DETECTION.
: STACK ( SIZEINCELLS -- )
  CREATE  HERE ,                        \ SET TO INDICATE EMPTY
          CELLS ALLOT ;

: EMPTY  ( &STACK -- )  DUP ! ;
: EMPTY? ( &STACK -- TRUTH )  DUP @ = ;

: PUSH  ( N &STACK -- )  1 CELLS OVER +!  @ ! ;
: 2PUSH ( D &STACK -- )  2 CELLS OVER +!  @ 2! ;

: POP  ( &STACK -- N )  DUP >R @ @   -1 CELLS R> +! ;
: 2POP ( &STACK -- D )  DUP >R @ 2@  -2 CELLS R> +! ;

                                        \ CREATES A NAMED ARRAY.
: ARRAY ( SIZEINCELLS -- )
        CREATE CELLS ALLOT
        DOES> ( U -- &ELEMENT ) SWAP CELLS + ;


\ FOSM  --  VARIABLES  &  <START  >END  C?  ?? -----------------------------

VARIABLE TEMP                           \ TO HOLD SINGLE CELLS TEMPORARILY.
VARIABLE THISCHAR
VARIABLE FIRSTCHAR
VARIABLE LASTCHAR
VARIABLE OLDDEPTH
VARIABLE CATCH#
VARIABLE THROW#

: <START  ( -- CHARSMATCHED )
        THISCHAR @  FIRSTCHAR @ - CHARS/ ;
: >END  ( -- CHARSUNMATCHED )
        LASTCHAR @  THISCHAR @ - CHARS/ ;


\ C?, ??, $=?, $ANY AND $NONE ARE THE 5 TEST WORDS.  IF THE TEST FAILS THEN
\ THE NEXT WORD ON THE DATA STACK IS EXECUTED.  THIS WILL BE ~FAIL IF
\ BACKTRACKING IS EXHAUSTED, WHICH DROPS RETURN ADDRESSES TO FORCE AN EXIT
\ FROM <<MATCH>> WITH JUST 0 LEFT.
\ THESE WORDS MUST NEVER BE USED WITHIN : .. ; ONLY WITHIN << .. >>.

\ IF THE TEST SUCCEEDS, THEN THE STRING POINTER IS INCREMENTED AND THE NEXT
\ CHARACTER FOUND.  A SUCCESSFUL TEST OF THE LAST CHARACTER WILL FIND THE
\ CHARACTER BEYOND THE STRING, WHICH IS HARMLESS.
: C?  ( XT TRUTH -- I*X )
        IF THISCHAR @ LASTCHAR @ U<     \ IF TEXT NOT EXHAUSTED...                                \ CMJ
          IF 1 CHARS THISCHAR +!
            THISCHAR @ C@ EXIT          \ CONTINUE WITH NEXT CHARACTER
        THEN THEN
        EXECUTE ;                       \ BACKTRACK

: ??  ( XT CHAR CONTINUE=FALSE -- I*X )
               ( CONTINUE=TRUE -- )
        IF EXIT THEN                    \ CONTINUE MATCHING
        DROP EXECUTE ;                  \ BACKTRACK

\ FOSM -- UPC  CH=  CASEON  CASEOFF  $=? -----------------------------------

: =?  POSTPONE = ; IMMEDIATE            \ USED TO COMPARE CHARACTERS

                                        \ RE-DEFINE =? TO ADD DEBUG
                                FD: VARIABLE =?S
                                FD: : =?  1 =?S +!  DUP EMIT = ;
                                FD: : .NEXT  CR <START 1+ SPACES ;

                                        \ CONVERT AN ASCII CHAR. TO UPPERCASE.
                                        \ NOTE: IF YOU HAVE AN ASSEMBLY CODE
                                        \ VERSION OF UPC, THEN USE IT INSTEAD
                                        \ FOR GREATER SPEED.
: UPC ( CHAR -- UPPERCASECHAR )
          DUP DUP 123 <  AND  96 >  32 AND  XOR ;
: (CH=) ( CH CH--F)
        UPC SWAP UPC =? ;

VARIABLE 'CH=
: CH= ( -- )                            \ MACRO AS AFFECTS PERFORMANCE
        S" 'CH= @ EXECUTE" EVALUATE ; IMMEDIATE
: CASEON ( -- )
        ['] =?  'CH= ! ;
: CASEOFF ( -- )
        ['] (CH=)  'CH= ! ;
CASEON                                  \ DEFAULT IS CASE SENSITIVE

\ FOSM -- $ANY $NONE $=? ---------------------------------------------------

: =ANYOF ( CHAR &STRING COUNT -- FLAG )
      >R 2>R 0 2R> R>                 \ BURY FLAG
      0
      DO  2DUP C@ =?
        IF  2>R NOT 2R>  LEAVE        \ SET BURIED FLAG
        THEN CHAR+
      LOOP 2DROP ;

                                \ USE AS << BRACKET    S" ()" $ANY >>
                                \ AND    << NOTBRACKET S" ()" $NONE >>
: $ANY ( -- POSTPONES C A L -- )      \ MATCH CHAR AGAINST ANY IN STRING
      S" =ANYOF C?"                   \ AND ADVANCE POINTER.
      EVALUATE ; IMMEDIATE

: $NONE ( -- POSTPONES C A L -- )     \ MATCH CHAR AGAINST ANY NOT IN STRING
      S" 2>R DUP 2R> =ANYOF NOT ??"   \ AND DO NOT ADVANCE POINTER.
      EVALUATE ; IMMEDIATE

: $=? ( XT CHAR &STRING COUNT -- XT CHARBEYOND )
      2>R DROP                        \ CHAR IS REDUNDANT
      -1  THISCHAR @ 2R>              \ BURY CONTINUE FLAG UNDER THISCHAR
      DUP >END >                      \ IF STRING IS EXHAUSTED...
      IF  2DROP  SWAP 0= SWAP         \ CANCEL CONTINUE FLAG
      ELSE  CHARS OVER + SWAP
        ?DO  DUP C@ I C@ CH= NOT      \ IF MISMATCH...
          IF  SWAP 0= SWAP LEAVE THEN \ CANCEL CONTINUE FLAG
        CHAR+
        1 CHARS +LOOP
      THEN SWAP
      IF DUP THISCHAR !  C@ EXIT THEN \ CONTINUE
      DROP EXECUTE ;                  \ BACKTRACK

\ FOSM -- =CHAR =RANGE ------------------------------------------------------

: =CHAR ( CHAR -- )
        CREATE IMMEDIATE  C,
        DOES> C@ S" LITERAL =? C?" EVALUATE ;

: =RANGE ( FROMCHAR TOCHAR -- )
        CREATE IMMEDIATE  1+ C,  C,            \ STORE TOCHAR, THEN FROMCHAR
        DOES> DUP C@  SWAP CHAR+ C@
              S" LITERAL LITERAL" EVALUATE
              FD: S" OVER EMIT [CHAR] - EMIT DUP 1- EMIT  1 =?S +!" EVALUATE
              S" WITHIN C?" EVALUATE ;

\ FOSM -- <<  >> -----------------------------------------------------------

: ~<< ( XT -- J*X )
        SKIPCALLER
        SKIPCALLER                      \ UN-NEST THE PATTERN
        EXECUTE ;                       \ THEN BACKTRACK

: ~>> ( XT RA -- J*X )
        SKIPCALLER
        RASTASH RAPUT RAUNSTASH         \ RE-NEST THE PATTERN AGAIN
        EXECUTE ;                       \ THEN BACKTRACK

: (<<) ( ~FAIL 1STCHAR -- ~FAIL ~<< 1STCHAR )
                                FD: ." <<" .NEXT
        ['] ~<<  SWAP ;

: (>>) ( ~FAIL ~<< ANYCHAR -- ~FAIL ANYCHAR )
       (       I*X ANYCHAR -- RA ~>> ANYCHAR )
        OVER ['] ~<< =                  \ IF ALL OPTIONS TAKEN
        IF NIP
        ELSE TEMP !                     \ SAVE ANYCHAR
          RASTASH RACOPY RAUNSTASH      \ SAVE RETURN ADDRESS FOR RE-NESTING
          ['] ~>>  TEMP @               \ RESTORE ANYCHAR
        THEN ;

\ << AND >> CREATE A PATTERN WORD WITH THE FOLLOWING STRUCTURE:
\       (:) (<<) ... (>>) (;)
\ NOTE: IF BACK-TRACKING IS TO WORK CORRECTLY, PATTERNS CAN BE USED WITHIN
\       PATTERNS, BUT NOT WITHIN : .. ; WORDS.

: << ( -- )
        :  POSTPONE (<<) ;
: >> ( -- )
        POSTPONE (>>)  POSTPONE ; ; IMMEDIATE

\ NEXT 2 LINES ARE FOR F-PC, WHERE : CALLS ] TO INTERPRET THE REST OF THE
\ DEFINITION, SO WE HAVE TO USE F-PC'S (:) INSTEAD.
FPC: : << ( -- )
FPC:        (:)  POSTPONE (<<) ] ;


\ FOR DEBUG, REDEFINE << TO ECHO THE PATTERN AT COMPILE TIME.
                              \ SINCE THE INPUT LINE WON'T CONTAIN ASCII 0,
                              \ 0 WORD CONSUMES THE WHOLE LINE.
FD: : <<
FD:      >IN @ >R CR CR ." << " 0 WORD COUNT TYPE R> >IN !
FD:      << ;

\ FOSM -- {  |  } {{ }} -----------------------------------------------------

512 2* STACK VSTACK                     \ A STACK TO STORE 512 PAIRS OF
                                        \ VARIABLE ADDRESSES AND VALUES.

: UNSTACK  ( &VSTACK -- )               \ REMOVES ALL VARIABLES STORED ON THE
  VSTACK @ SWAP ?DO                     \ VSTACK AND SETS THEM.
    VSTACK 2POP !
  2 CELLS +LOOP ;

VARIABLE '~{                            \ VECTOR FOR ~{.  ~{ IS TEMPORARILY
                                        \ RE-VECTORED BY FALL BELOW.

: (~{) ( RA THISCHAR VSTACK -- CHAR OPTIONFLAG )
        SKIPCALLER
        UNSTACK
        THISCHAR !
        RAGET RADROP RAPUT               \ JUMP BACK TO OPTIONAL PATH
                                FD: ."  - BACKTRACK TO {" .NEXT
        THISCHAR @ C@
        FALSE ;                          \ ON RE-TRY, MATCH ALTERNATE.  \ CMJ

: ~{  ( S--S)
        SKIPCALLER  '~{ @ EXECUTE ;     \ RESUME AFTER BACKTRACKING

' (~{) '~{ !

: ({) ( CHAR -- RA THISCHAR VSTACK ~{ CHAR OPTIONFLAG )
        TEMP !
        RACOPY
        THISCHAR @  VSTACK @  ['] ~{  TEMP @
        TRUE ;                          \ INITIALLY MATCH THE OPTION.   \ CMJ

\ WHEN IMPLEMENTED USING ANS FORTH CONTROL FLOW WORDS, THE ORIGINAL STRUCTURE
\ { .. }} IS NOT COMPATIBLE WITH { .. } SO IT HAS BEEN REPLACED BY
\ {{ .. }}.

                                FD: : .{  ." {"  .NEXT ;  : .}  ." }"  .NEXT ;
                                FD: : .|  ." |"  .NEXT ;
                                FD: : .{{ ." {{" .NEXT ;  : .}} ." }}" .NEXT ;

: { ( ADDR -- )                 FD: POSTPONE .{
        POSTPONE ({)  POSTPONE IF ; IMMEDIATE

: | ( ADDR -- ADDR )
        POSTPONE ELSE           FD: POSTPONE .|
        ; IMMEDIATE

: }  ( -- ADDR )
        POSTPONE THEN           FD: POSTPONE .}
        ; IMMEDIATE

: {{ ( ADDR -- )                FD: POSTPONE .{{
        POSTPONE BEGIN  POSTPONE {  1 CS-ROLL ; IMMEDIATE

: }} ( ADDR -- ADDR )           FD: POSTPONE .}}
        POSTPONE AGAIN  POSTPONE } ; IMMEDIATE

\ LAZY VERSIONS
                                FD: : .{L  ." {L"  .NEXT ;
                                FD: : .{{L ." {{L" .NEXT ;

: {L ( ADDR -- )                FD: POSTPONE .{L
        POSTPONE ({)  POSTPONE 0= POSTPONE IF ; IMMEDIATE

: {{L ( ADDR -- )                FD: POSTPONE .{{L
        POSTPONE BEGIN  POSTPONE {L  1 CS-ROLL ; IMMEDIATE

\ FOSM -- *{  }* ------------------------------------------------------------

\ *{ .. }* LOOPS UNTIL THE COUNT DECREMENTS TO ZERO.  THE LOOP COUNT IS KEPT
\ ON THE RETURN STACK BECAUSE ALTERNATIVE PATHS WITHIN *{ .. }* MAY BE LEFT
\ ON THE DATA STACK BY { OR {{.

: ~*{  ( XT  -- I*X )
        SKIPCALLER
        RASTASH  R> DROP  RAUNSTASH     \ DROP COUNT FROM UNDER RETURN STACK
        EXECUTE ;                       \ BACKTRACK

: ~}*  ( XT COUNT -- I*X )
        SKIPCALLER
        RASTASH  >R  RAUNSTASH          \ BURY COUNT UNDER RETURN STACK
        EXECUTE ;                       \ BACKTRACK

: (*{)  ( CHAR COUNT -- ~*{ CHAR )
        RASTASH  >R  RAUNSTASH          \ BURY COUNT UNDER RETURN STACK
        ['] ~*{  SWAP ;

: (}*)  ( ~*{ CHAR -- COUNT=0 CHAR FALSE )               \ CONTINUE PROCESSING
        (          -- COUNT>0 CHAR COUNT-1 TRUE )        \ BRANCH BACK TO *{
        (  XT CHAR -- XT COUNT=0 ~}* CHAR FALSE )        \ CONTINUE PROCESSING
        (          -- XT COUNT>0 ~}* CHAR COUNT-1 TRUE ) \ BRANCH BACK TO *{
        RASTASH  R> TEMP !  RAUNSTASH   \ RESTORE COUNT FROM RETURN STACK
        OVER ['] ~*{ =
        IF NIP                          ( -- CHAR )
        ELSE TEMP @  ['] ~}*  ROT       ( -- XT COUNT ~}* CHAR )
        THEN
        TEMP @ 1- ?DUP 0= ;

                                FD: : .*{ ." *{" .NEXT ;  : .}* ." }*" .NEXT ;

: *{  ( -- ADDR )               FD: POSTPONE .*{
        POSTPONE BEGIN  POSTPONE (*{) ; IMMEDIATE
: }*  ( ADDR -- )               FD: POSTPONE .}*
        POSTPONE (}*)  POSTPONE UNTIL ; IMMEDIATE

\ FOSM -- !{   }! ----------------------------------------------------------

VARIABLE HOLE
                                        \ REPLACES THE &VARIABLE SAVED ON
                                        \ THE VSTACK WITH &HOLE.  THIS STOPS
                                        \ THE VARIABLE BEING CHANGED BY UNSTACK
: HIDEVAR ( &VARIABLE &STACK -- )
  CELL+ CELL+  VSTACK @ CELL+ CELL+  SWAP
  ?DO
    I @ OVER = IF
      HOLE I !
    THEN
  2 CELLS +LOOP  DROP ;

: ~!{  ( XT -- I*X )
        SKIPCALLER  UNSTACK  RASTASH
        R> R>  !                        \ RESTORE VARIABLE FROM RETURN STACK
        RAUNSTASH  EXECUTE ;            \ BACKTRACK

: ~}!  ( XT CHAR -- N &VARIABLE I*X )
        SKIPCALLER  RASTASH
        DUP >R  DUP @ >R  !             \ SAVE THE VARIABLE TO RETURN STACK
        RAUNSTASH  EXECUTE ;            \ BACKTRACK

: !{  ( CHAR N &VARIABLE -- &TOPOFVSTACK ~!{ CHAR )
        RASTASH
        DUP >R  DUP @ >R                \ SAVE &VARIABLE AND DEFAULT VALUE
        !                               \ SET VARIABLE TO N
        RAUNSTASH
        VSTACK @  ['] ~!{  ROT ;

: }!  ( &TOPOFVSTACK ~!{ CHAR -- CHAR )
      ( XT CHAR -- N &VARIABLE )
        RASTASH
        OVER ['] ~!{ =
        IF  NIP
          R> R@ !  R>                   \ RESTORE &VARIABLE AND INITIAL VALUE
          ROT HIDEVAR
        ELSE  R> R>
          DUP @ OVER  2SWAP !
          ROT  ['] ~}! SWAP
        THEN RAUNSTASH ;

\ FOSM -- <<MATCH>> --------------------------------------------------------

1024 STACK SAVED
VARIABLE NOTES                                                          \ CMJ

: ~FAIL ( -- 0 )
        SKIPCALLER SKIPCALLER  0 ;

: INITVARS  ( $ADDR COUNT -- )
        OVER DUP  FIRSTCHAR !  THISCHAR !  CHARS
        + LASTCHAR !
        DEPTH OLDDEPTH !  VSTACK EMPTY  SAVED EMPTY
        0 CATCH# !
        0 NOTES !                                                       \ CMJ
        ['] (~{) '~{ ! ;                                                \ CMJ

: CLEANUP  ( S -- )
        DEPTH OLDDEPTH @ DO  SAVED PUSH  LOOP ;

: <<MATCH>>  ( $ADDR COUNT 'PATTERN -- MATCHFOUND? )
        >R  INITVARS  ['] ~FAIL  THISCHAR @ C@  R>
        EXECUTE                         \ EXECUTE THE PATTERN.
        CLEANUP -1 ;                    \ ~FAIL WILL SKIP THIS.

\ FOSM -- END  ANY  SPC  -TRAIL  --------------------------------------------
                                FD: : .END ." END" .NEXT ;
: END  ( -- )                   FD: POSTPONE .END
        S" >END 1 < ?? " EVALUATE ; IMMEDIATE   \ MORE EFFICIENT THAN JUST
                                                \ << END >END 1 < ?? >>

                                FD: : .ANY ." ANY" .NEXT ;
: ANY  ( -- )                   FD: POSTPONE .ANY
        S" DROP TRUE C?" EVALUATE ; IMMEDIATE   \ MORE EFFICIENT THAN JUST
                                                \ << ANY DROP TRUE C? >>
BL =CHAR SPC
<< -TRAIL  {{ SPC }} END >>

\ FOSM -- &{  }{  }& ------------------------------------------------------

VARIABLE (HEAD)
VARIABLE (TAIL)

: NOTEHEAD  ( --ADDR ADDR)
        THISCHAR @ (HEAD) ;
: NOTETAIL  ( --N ADDR)
        <START (TAIL) ;
: RESETTHIS  ( -- )
        DROP  (HEAD) @  DUP C@ SWAP  THISCHAR ! ;

<< LINEUP  ( -- )               FD: ." LINEUP" .NEXT
        <START (TAIL) @ = ?? >>


                                FD: : .&{ ." &{" .NEXT ;  : .}{ ." }{" .NEXT ;
                                FD: : .}& ." }&" .NEXT ;

: &{  ( -- )
        POSTPONE NOTEHEAD  POSTPONE !{
                                FD: POSTPONE .&{
        ; IMMEDIATE
: }{  ( -- )
        POSTPONE NOTETAIL  POSTPONE !{  POSTPONE RESETTHIS
                                FD: POSTPONE .}{
        ; IMMEDIATE
: }&  ( -- )                    FD: POSTPONE .}&
        POSTPONE LINEUP  POSTPONE }!  POSTPONE }!
        ; IMMEDIATE

\ FOSM  --  {!}  NOTES  NOTE  NOTE@  SYNOPSIS  >N N@ N> ---------------------

: {!}  ( S N ADDR -- S )
        DUP @ OVER  VSTACK 2PUSH ! ;

1024 CONSTANT MAXNOTES
MAXNOTES ARRAY SYNOPSIS

: NOTE  ( NOTE -- )                             \ SAVE A NOTE IN SYNOPSIS
        NOTES @  TUCK SYNOPSIS !  1+ NOTES {!} ;

: NOTE@  ( INDEX -- NOTE )                      \ EXTRACT A NOTE FROM SYNOPSIS
        SYNOPSIS @ ;

: >N    ( NOTE -- )                             \ ALIAS FOR NOTE
        POSTPONE NOTE ; IMMEDIATE

: N@    ( -- NOTE )                             \ GET LATEST NOTE
        NOTES @ 1-  SYNOPSIS @ ;

: N>    ( -- NOTE )                             \ COMPLEMENT OF >N
        N@  NOTES @ 1- NOTES {!} ;

\ FOSM  --  FAIL ABANDON CUT{ }CUT <<ANOTHER>> -----------------------------

: FAIL  ( -- )                                                          \ CMJ
        S" 0 ??" EVALUATE ; IMMEDIATE
: CATCH{  ( -- )
        S" CATCH#  !{" EVALUATE ; IMMEDIATE
: }CATCH  ( -- )
        S" }!" EVALUATE ; IMMEDIATE
: ~FALL ( RA THISCHAR VSTACK -- )
        SKIPCALLER
        CATCH# @  THROW# @ =
        IF  ['] (~{) DUP '~{ !          \ RESTORE NORMAL BACKTRACKING
                                        FD: ." CANCEL FALLING THROUGH" .NEXT
        ELSE UNSTACK DROP RADROP        FD: ." FALL THROUGH" .NEXT ." {" .NEXT
        THEN
        EXECUTE ;                       \ BACKTRACK

: FALL  ( LABEL -- )                    FD: ." SET BACKTRACK TO FALL THROUGH" .NEXT
        THROW# !  ['] ~FALL '~{ ! ;     \ SET TO FALL THROUGH ALTERNATIVES
: <THROW  ( -- )
        S" FALL FAIL" EVALUATE ; IMMEDIATE

VARIABLE LABEL#  0 LABEL# !

: LABEL  ( -- )
        1 LABEL# +!  LABEL# @ CONSTANT ;

LABEL ABANDONLABEL
: ABANDON  ( -- )                                                        \ CMJ
        S" ABANDONLABEL <THROW" EVALUATE ; IMMEDIATE

: LOCAL  ( -- LABEL )
        LABEL# @ ;
: LABEL+  ( -- LABEL &LABEL )
        LOCAL 1+  LABEL# ;
: LOCAL{  ( -- )
        S" LABEL+ !{" EVALUATE ; IMMEDIATE
: }LOCAL  ( -- )
        S" }!" EVALUATE ; IMMEDIATE

: {CATCH}  ( -- )
        S" CATCH{ { | FAIL } }CATCH" EVALUATE ; IMMEDIATE              \ CMJ

VARIABLE TMP
: THROWTMP  ( -- )
     S" { | TMP @ <THROW }" EVALUATE ; IMMEDIATE                       \ CMJ
: {THROW}  ( LABEL -- )
        S" TMP !{ THROWTMP }!" EVALUATE ; IMMEDIATE

: CUT{  ( -- )
        S" LOCAL{ LOCAL {CATCH}" EVALUATE ; IMMEDIATE
: }CUT  ( -- )
        S" LOCAL {THROW} }LOCAL" EVALUATE ; IMMEDIATE

: EXCISE (  I*X CHAR -- CHAR )          FD: ." EXCISE" .NEXT
       >R 4 0 DO 2DROP LOOP RADROP R> ;

: /{    ( -- )
     S" CUT{ { " EVALUATE ; IMMEDIATE
: }/    ( -- )
     S" LOCAL <THROW | EXCISE } }LOCAL" EVALUATE ; IMMEDIATE

LABEL -MTCH
: //{   ( -- )
        S" -MTCH {CATCH} { CUT{ -MTCH {THROW}" EVALUATE ; IMMEDIATE
: }//   ( -- )
        S" }CUT FAIL | EXCISE }" EVALUATE ; IMMEDIATE


: <<ANOTHER>> ( -- MATCHFOUND? )
        SAVED EMPTY?
        IF 0
        ELSE DEPTH OLDDEPTH !
          ['] (~{) '~{ !                                                \ CMJ
          BEGIN  SAVED POP SAVED EMPTY? UNTIL
          FAIL CLEANUP -1  THEN ;

CR .( ESSENTIAL FOSM COMPILED )

\ OPTIONAL FOSM  --  $ $- --------------------------------------------- \ CMJ
CR .( COMPILING OPTIONAL FOSM )

\ NOTE: THE FOLLOWING 2 WORDS ARE GENERALLY USEFUL WITHIN FOSM AS THEY AVOID
\       CREATING NAMED CONSTANTS USING =CHAR AND =RANGE.

                                        \ MACRO TO PULL A CHARACTER FROM THE
                                        \ INPUT STREAM AND OUTPUT TEXT TO
                                        \ COMPILE A MATCH WITH IT, SIMILAR TO
                                        \ =CHAR.  EG "$ A".
                                        \ USE ONLY WITHIN A PATTERN.
: $  ( -- )
        BL WORD COUNT 0= ABORT" $ EXPECTS A CHARACTER"
        C@  S" LITERAL CH= C?" EVALUATE ; IMMEDIATE

                                        \ MACRO TO PULL 3 CHARACTERS FROM THE
                                        \ INPUT STREAM AND OUTPUT TEXT TO
                                        \ COMPILE A MATCH WITH THEM, SIMILAR TO
                                        \ TO =RANGE. EG "$- 0-9".
                                        \ USE ONLY WITHIN A PATTERN.
: $-  ( -- )
        BL WORD COUNT
        3 = NOT ABORT" $- EXPECTS 3 CHARACTERS"
        DUP C@  SWAP CHAR+ CHAR+ C@ 1+  SWAP
              S" LITERAL LITERAL" EVALUATE
        FD: S" OVER EMIT [CHAR] - EMIT DUP 1- EMIT  1 =?S +!" EVALUATE
              S" WITHIN C?" EVALUATE ; IMMEDIATE

                                \ FASTER THAN USING $ANY AND $NONE AS LOOK UP
                                \ IS MORE EFFICIENT.  USE AS:
                                \   : ()STRING  S" ()" ;
                                \   ()STRING ANYOF ANY()
                                \ TO CREATE A TABLE ANY() MATCHING ANY OF "("
                                \ OR ")" FOR USE AS:
                                \   << BRACKET  ANY() C? >>
                                \   << NOTBRACKET  DUP ANY() NOT ?? >>
VARIABLE TEMPTABLE

: CREATETABLE (  -- )
        CREATE HERE TEMPTABLE !
        128 CHARS ALLOT
        TEMPTABLE @ 128 0 FILL ;

: FILLTABLE ( &STRING COUNT -- )
        0
        DO TRUE OVER C@                 \ GET CHAR FROM STRING
          CHARS TEMPTABLE @ + C!        \ SET MATCHING ENTRY IN TABLE TO TRUE
          CHAR+
        LOOP DROP ;

: ANYOF ( C: &STRING COUNT -- )
        ( R: CHAR -- ISMEMBER )
        CREATETABLE
        FILLTABLE
  DOES> SWAP CHARS + C@ ;               \ LOOK UP CHAR IN TABLE

                                \ MORE CONVENIENT THAN ANYOF IS ANYOF$:
                                \   ANYOF$ ANY() ()
                                \ TO CREATE A TABLE ANY() MATCHING ANY OF
                                \ "(" OR ")" FOR USE AS:
                                \   << BRACKET  ANY() C? >>
                                \   << NOTBRACKET  DUP ANY() NOT ?? >>
: ANYOF$ ( R: CHAR -- ISMEMBER )
        CREATETABLE
        BL WORD COUNT
        FILLTABLE
  DOES> SWAP CHARS + C@ ;

\ OPTIONAL FOSM  --  C{  }C{  }C ---------------------------------------------

VARIABLE FROM
VARIABLE CNT

: 0CNT  ( -- 0 ADDR)
        0 CNT ;
: INC  ( -- )
        CNT @ 1+  CNT {!} ;

<< OK?  CNT @  FROM  @  < ?? >>
<< DONE?  CNT @ FROM  @ = ?? >>

: C{  ( -- )
        S" FROM !{ 0CNT !{ { OK?"  EVALUATE ;  IMMEDIATE         \ CMJ
: }C{  ( -- )
        S" INC } { OK?"  EVALUATE ;  IMMEDIATE                   \ CMJ
: }C  ( -- )
        S" INC } DONE? }! }!"  EVALUATE ; IMMEDIATE

\ OPTIONAL FOSM  --  P{  }P{  }P --------------------------------------------

VARIABLE FLAGS
: USED  ( S -- S )
       FLAGS @ OR  FLAGS {!} ;

: USED?  ( N -- F )
       FLAGS @ AND 0= ;

VARIABLE CLAUSES
: >BIT  ( N--N)
       1 SWAP 1- 0 ?DO 2* LOOP ;
                                                                                                                                 
: P{  ( -- )
      CLAUSES @  1 CLAUSES !
      S" FROM !{ 0 FLAGS !{  FROM @ *{ { 1 USED? ??" EVALUATE ; IMMEDIATE
: }P{  ( -- )
        CLAUSES @ >BIT  S" LITERAL  USED  |  {" EVALUATE
        1 CLAUSES +!
        CLAUSES @ >BIT  S" LITERAL USED? ??" EVALUATE ;                                                                 IMMEDIATE

: }P  ( -- )
        CLAUSES @ >BIT  S" LITERAL USED | FAIL" EVALUATE
        CLAUSES @ 0 DO [COMPILE] } LOOP
        S" }* }! }!" EVALUATE
        CLAUSES ! ; IMMEDIATE

CR .( OPTIONAL FOSM COMPILED )

\ FOSM TESTS ---------------------------------------------------------------

FD: CR .( EXECUTING FOSM TESTS )

\ THE TESTS ARE DESIGNED TO PASS THROUGH WHEN RESULTS MATCH EXPECTATIONS, BUT
\ REPORT AND STOP AT ANY THAT ARE INCORRECT.
FD: : RETURN  ( -- )
FD:     ."  RETURN" KEY 13 = IF ( DO NOTHING ) ELSE ABORT THEN ;
FD: : CONTINUE  ( -- )
FD:     CR ." TO CONTINUE, PRESS" RETURN ;
FD: : CHECK  ( RESULTFLAG EXPECTEDFLAG -- )
FD:      = IF ."  OK" ELSE ."  ERROR" CONTINUE THEN ;
FD: : PASSED?  ( RESULTFLAG  -- )
FD:        TRUE CHECK ;
FD: : FAILED?  ( RESULTFLAG -- )
FD:        FALSE CHECK ;
FD: : TEST ( -- MATCHFLAG )
FD:        0 =?S !
FD:        ' [CHAR] # WORD COUNT
FD:        CR [CHAR] " EMIT 2DUP TYPE [CHAR] " EMIT CR SPACE ROT <<MATCH>>
FD:        DUP NOT ."  => " IF ." NO MATCH," ELSE ." MATCH," THEN
FD:        ."  CHARS MATCHED = " =?S @ . ;

\ CHARACTER TEST PATTERNS
FD: CHAR A =CHAR A  CHAR B =CHAR B  CHAR C =CHAR C  CHAR D =CHAR D
FD: CHAR E =CHAR E

\ RANGE TEST PATTERNS
FD: CHAR A CHAR B =RANGE A-B

\ FOSM TESTS BEGIN HERE -----------------------------------------------------

FD: << TCHAR  A >>
FD: TEST TCHAR A# PASSED?               TEST TCHAR B# FAILED?

FD: << TRANGE  A-B >>
FD: TEST TRANGE A# PASSED?              TEST TRANGE C# FAILED?

FD: << TANYOF  S" {}" $ANY >>
FD: TEST TANYOF {# PASSED?              TEST TANYOF A# FAILED?

FD: << TNONEOF  S" {}" $NONE >>
FD: TEST TNONEOF A# PASSED?              TEST TNONEOF {# FAILED?

FD: << TSTRING  S" ABC" $=? >>
FD: TEST TSTRING ABC# PASSED?           TEST TSTRING ABA# FAILED?
FD: TEST TSTRING AB# FAILED?

FD: << TOPTIONAL  { A } B >>
FD: TEST TOPTIONAL B# PASSED?           TEST TOPTIONAL AB# PASSED?
FD: TEST TOPTIONAL A# FAILED?

FD: << TLAZYOPTIONAL  {L A } B >>
FD: TEST TLAZYOPTIONAL B# PASSED?       TEST TLAZYOPTIONAL AB# PASSED?
FD: TEST TLAZYOPTIONAL A# FAILED?

FD: << TALTERNATE  { A | B } >>
FD: TEST TALTERNATE A# PASSED?          TEST TALTERNATE B# PASSED?

FD: << TNEST  TOPTIONAL TCHAR >>
FD: TEST TNEST ABA# PASSED?             TEST TNEST A# FAILED?

FD: << TMANY  A {{ B }} C >>
FD: TEST TMANY ABBC# PASSED?            TEST TMANY AC# PASSED?

FD: << TLAZYMANY  A {{L B }} C >>
FD: TEST TLAZYMANY ABBC# PASSED?        TEST TLAZYMANY AC# PASSED?

FD: << TMULTIPLE  3 *{ A }* >>
FD: TEST TMULTIPLE AAA# PASSED?         TEST TMULTIPLE AA# FAILED?

FD: VARIABLE N
FD: << TVARIABLE 65 N !{ N @ =? C? }! >>
FD: TEST TVARIABLE A# PASSED?

FD: << TEND  { A B } END >>
FD: TEST TEND AB# PASSED?               TEST TEND BA# FAILED?

FD: << TAND  &{ A B A }{ 3 *{ { ANY } }* }& >>
FD: TEST TAND ABA# PASSED?              TEST TAND ABC# FAILED?

FD: << TNOT  /{ { B } END }/ B B END >>
FD: TEST TNOT BBB# FAILED?              TEST TNOT BB# PASSED?

FD: << TABANDON  A { END | ABANDON } >>
FD: TEST TABANDON A# PASSED?            TEST TABANDON AB# FAILED?

FD: << TCUT  { A B | A } CUT{ { B | B } }CUT C >>
FD: TEST TCUT ABC# PASSED?              TEST TCUT ABBC# PASSED?

FD: : {}STRING S" {}" ;  {}STRING ANYOF {}
FD: << TANYOF2  {} C? >>
FD: TEST TANYOF2 {# PASSED?             TEST TANYOF2 A# FAILED?

FD: << TNONEOF2  DUP {} NOT ?? >>
FD: TEST TNONEOF2 A# PASSED?             TEST TNONEOF2 {# FAILED?

FD: << TCOMBS  3 C{ A }C{ B }C{ C }C{ D }C{ E }C >>
FD: TEST TCOMBS ACD# PASSED?            TEST TCOMBS AAA# FAILED?

FD: << TPERMS  3 P{ A }P{ B }P{ C }P{ D }P{ E }P >>
FD: TEST TPERMS CBAD# PASSED?           TEST TPERMS BA# FAILED?

FD: CR .( FOSM TESTS COMPLETED )

\ APPENDIX ------------------------------------------------------------------

\ CHANGES FROM EUROFORML VERSION
\ ==============================
\ WHERE THE CODE GIVEN HERE DIFFERS FROM EUROFORML, SOME CHANGES WERE MADE
\ FOR ANS COMPATIBILITY.  OTHER CHANGES ARE MARKED WITH THE COMMENT \ CMJ.
\ THESE ARE:
\ 1. THE SYNTAX { .. | .. } AND { .. | .. }} CANNOT BE ACHIEVED PORTABLY AND
\    HAS BEEN REPLACED BY { .. | .. } AND {{ .. }}.
\ 2. GREEDY EVALUATION IS NOW THE DEFAULT.
\    LAZY EVALUATION (SIGNIFIED BY L) IS ALSO AVAILABLE.
\    THE OPTION { .. } IS MATCHED FIRST, SKIPPED ON RE-TRY.
\    {L .. } REVERSES THIS.
\    {{ .. }} MATCHES AS MANY TIMES AS POSSIBLE.
\    {{L .. }} MATCHES AS FEW TIMES.
\ 3. DOES<< .. >>DOES HAS BEEN DROPPED IN FAVOUR OF MACROS WHICH ARE MORE
\    EFFICIENT.
\ 4. ABANDON IS NOW A PATTERN, NOT A LABEL.
\ 5. OPTIONAL DEBUG FACILITIES (SEE FOSMDEBUG) PRINT THE PROGRESS OF
\    MATCHING.
\ 6. OPTIONAL WORDS $ $- PROVIDE ANONYMOUS PATTERNS.
\ 7. OPTIONAL TEST WORDS (SEE END OF FILE) CHECK THE CORRECT OPERATION OF
\    FOSM.


\ SUGGESTIONS FOR THE NEXT VERSION
\ ================================
\ ADD A 'POSITIVE CLOSURE' CONSTRUCT:
\ {{ .. }} IS KNOWN AS A '*' KLEENE CLOSURE, MATCHING 0 OR MORE OCCURENCES.
\ X {{ X }} IS KNOWN AS A '+' POSITIVE CLOSURE, MATCHING 1 OR MORE.
\
\ PROVIDE A MACRO FOR THE COMMON CONSTRUCTION:
\       CUT{ N *{ { .. } }* }CUT
\
\ ADD COMPILER SECURITY TO MATCH THE PAIRS: << { {{ !{ /{ &{ CUT{
\ ADD COMPILER SECURITY FOR USE OUTSIDE PATTERNS.
\ ADD RUN-TIME CHECKS FOR STACK OVER/UNDERFLOW.
\
\ CONSIDER MOVING 0 NOTES ! INTO INITVARS TO SIMPLIFY SETTING UP.
\
\ CONSIDER A FASTER VERSION OF UPC.
\
\ COMMENTS FROM BARRIE STOTT:
\ REMOVE DUPLICATE CODE IN TUTORIAL
\ EXTEND FD: TO TURN OFF MESSAGES, TO TURN OFF TESTS
\ MATCHING AGAINST INPUT STREAM, INCLUDING REFILL
\ AN OVERVIEW INCLUDING 'USEFUL FOR' AND 'NOT USEFUL FOR'

CR .( Size of FOSM1 is ) HERE SWAP - . CR