/****************************************************************************
*                                                                           *
*                           cryptlib MD5 Hash Routines                      *
*                       Copyright Peter Gutmann 1992-1996                   *
*                                                                           *
****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "crypt.h"
#include "cryptctx.h"
#include "md5.h"

/****************************************************************************
*                                                                           *
*                           Init/Shutdown Routines                          *
*                                                                           *
****************************************************************************/

/* Perform auxiliary init and shutdown actions on an encryption context */

#include <stdio.h>

int CALLBACK md5Init(CRYPT_INFO *cryptInfo){

    int res;
/*
    if ((res=krnlMemalloc(&cryptInfo,sizeof(CRYPT_INFO)))!=CRYPT_OK){
        printf("error1 %d\n",res);
        return res;
    }
*/
    if( ( res = krnlMemalloc( &cryptInfo->ctxHash.hashInfo,
                                 sizeof( MD5_CTX ) ) ) != CRYPT_OK ){
        return res;
    }

    MD5_Init( ( MD5_CTX * ) cryptInfo->ctxHash.hashInfo );

    return CRYPT_OK;
}

int CALLBACK md5End( CRYPT_INFO *cryptInfo )
    {
    /* Free any allocated memory */
    krnlMemfree( &cryptInfo->ctxHash.hashInfo );

    return( CRYPT_OK );
    }

/****************************************************************************
*                                                                           *
*                               MD5 Hash Routines                           *
*                                                                           *
****************************************************************************/

/* Hash data using MD5 */

int CALLBACK md5Hash( CRYPT_INFO *cryptInfo, BYTE *buffer, int noBytes )
    {
    MD5_CTX *md5Info = ( MD5_CTX * ) cryptInfo->ctxHash.hashInfo;

    /* If we've already called MD5_Final(), we can't continue */
    if( cryptInfo->ctxHash.done )
        return( CRYPT_ERROR_COMPLETE );

    if( !noBytes )
        {
        MD5_Final( cryptInfo->ctxHash.hash, md5Info );
        cryptInfo->ctxHash.done = TRUE;
        }
    else
        MD5_Update( md5Info, buffer, noBytes );

    return( CRYPT_OK );
    }

/* Internal API: Hash a single block of memory without the overhead of
   creating an encryption context */

void CALLBACK MD5Init(MD5_CTX * md5Info ){ 
    MD5_Init(md5Info); 
}
void CALLBACK MD5Update(MD5_CTX * md5Info, const BYTE * inBuffer, const int length ){
    MD5_Update(md5Info, (BYTE *)inBuffer, length);
}
void CALLBACK MD5Final(MD5_CTX * md5Info, BYTE * outBuffer){
    MD5_Final(outBuffer, md5Info );
}

void CALLBACK MD5Full(const BYTE * inBuffer, const int length, BYTE *outBuffer){
    MD5_CTX *md5Info, md5InfoBuffer;

    md5Info = &md5InfoBuffer;

    MD5_Init( md5Info );
    MD5_Update( md5Info, ( BYTE * ) inBuffer, length );
    MD5_Final( outBuffer, md5Info );
}


void CALLBACK md5HashBuffer( void *hashInfo, BYTE *outBuffer, const BYTE *inBuffer,
                    const int length, const HASH_STATE hashState )
    {
    MD5_CTX *md5Info = ( MD5_CTX * ) hashInfo, md5InfoBuffer;

    /* If the user has left it up to us to allocate the hash context buffer,
       use the internal buffer */
    if( md5Info == NULL )
        md5Info = &md5InfoBuffer;

    if( hashState == HASH_ALL )
        {
        MD5_Init( md5Info );
        MD5_Update( md5Info, ( BYTE * ) inBuffer, length );
        MD5_Final( outBuffer, md5Info );
        }
    else
        switch( hashState )
            {
            case HASH_START:
                MD5_Init( md5Info );
                /* Drop through */

            case HASH_CONTINUE:
                MD5_Update( md5Info, ( BYTE * ) inBuffer, length );
                break;

            case HASH_END:
                if( inBuffer != NULL )
                    MD5_Update( md5Info, ( BYTE * ) inBuffer, length );
                MD5_Final( outBuffer, md5Info );
            }

    /* Clean up */
    zeroise( &md5InfoBuffer, sizeof( MD5_CTX ) );
    }


BOOL WINAPI DllMain(
    HINSTANCE hinstDLL,  // handle to DLL module
    DWORD fdwReason,     // reason for calling function
    LPVOID lpReserved )  // reserved
{
    // Perform actions based on the reason for calling.
    switch( fdwReason ) 
    { 
        case DLL_PROCESS_ATTACH:
         // Initialize once for each new process.
         // Return FALSE to fail DLL load.
            break;

        case DLL_THREAD_ATTACH:
         // Do thread-specific initialization.
            break;

        case DLL_THREAD_DETACH:
         // Do thread-specific cleanup.
            break;

        case DLL_PROCESS_DETACH:
         // Perform any necessary cleanup.
            break;
    }
    return TRUE;  // Successful DLL_PROCESS_ATTACH.
}


int krnlMemalloc( void **pointer, int size ){
    *pointer=HeapAlloc(GetProcessHeap(),0,size);
    return( CRYPT_OK );
}

/* A safe free function which scrubs memory and zeroes the pointer.

    "You will softly and suddenly vanish away
     And never be met with again"   - Lewis Carroll,
                                      "The Hunting of the Snark" */

void krnlMemfree( void **pointer ){
    HeapFree(GetProcessHeap(),0,*pointer);
}

int CALLBACK md5strsize(){
    return sizeof(CRYPT_INFO);
}

int CALLBACK md5ctxsize(){
    return sizeof(MD5_CTX);
}