
0 [IF]

This is ANS FORTH version of new proposed SHA-256
cryptographic hash functions.  It taks 512-bit
input blocks and creates a 25-bit hash of it.
See specification on NIST website at:
http://csrc.nist.gov/encryption/tkhash.html

On my 600 MHZ Atholon PC TEST3 takes between
28.6-29.0 secs.

Enjoy

Jabari Zakiya

[THEN]

REQUIRE MARKER lib/ext/core-ext.f
REQUIRE /STRING lib/ext/string.f
REQUIRE D2* lib/ext/double.f

: 3DROP 2DROP DROP ;

\ ANS Forth code for proposed Secure Hash Algorithm 256 (SHA-256)
\ NIST spec at:  http://csrc.nist.gov/encryption/tkhash.html
\ For Little/Big Endian byte addressable CPUs, e.g. Pentium/Power PC
\ DEPENDENCIES: CORE EXT WORDSET ; COMMON USAGE 3DROP ?DO CELL-
\ Use of this code is free subject to acknowledgment of copyright.
\ Copyright (c) 2001 Jabari Zakiya -- jzakiya@mail.com  5/12/2001

\ ======================= MACRO Wordset =======================
\ MACRO wordset from Wil Baden's Tool Belt series in
\ Forth Dimensions (FD) Vol. 19, No. 2, July/August 1997
\ Original code has been modified to make more efficient
\ MACRO allows insertion of parameters following the macro
\ "\" represents place where parameter is inserted
\ Example:  MACRO  ??  " IF  \  THEN "
\ : FOO .. ?? EXIT .... ;  ?? compiles to -- IF EXIT THEN

\ PLACE and STRING for system if needed
\ Not needed for SwiftForth v 2.00.3, needed for Win32Forth V 4.10
\ : PLACE  ( caddr n addr -)  2DUP  C!  CHAR+  SWAP  CHARS  MOVE ;
: STRING ( char "ccc" -) WORD COUNT HERE OVER 1+ CHARS ALLOT PLACE ;

\ Versions of /STRING and ANEW if system doesn't have them
\ : /STRING ( a n k - a+k n-k) ( OVER MIN) TUCK - >R CHARS + R> ;
: ANEW  >IN @ BL WORD FIND IF EXECUTE ELSE DROP THEN >IN ! MARKER ;

: split-at-char  ( a  n  char  -  a  k  a+k  n-k)
  >R  2DUP  BEGIN  DUP  WHILE  OVER  C@  R@  -
            WHILE  1 /STRING  REPEAT  THEN
            R> DROP  TUCK  2>R  -  2R>
;

: DOES>MACRO  \ Compile the macro, including external parameters
  DOES> COUNT  BEGIN [CHAR]  \ split-at-char  2>R  EVALUATE  R@
               WHILE BL WORD COUNT EVALUATE 2R>  1 /STRING REPEAT
               R> DROP   R> DROP
;

\ Macro creation word which allows parameter insertion
: MACRO  CREATE  IMMEDIATE  CHAR  STRING  DOES>MACRO  ;

\ ====================  Start SHA-256 Code ====================
  DECIMAL
  32 CONSTANT CELLSIZE    \ CPU bitsize

  VARIABLE  SHAlen              \ Holds byte length of string
  CREATE SHAval  8 CELLS ALLOT  \ Holds hash after each block
  CREATE SHAsh  72 CELLS ALLOT  \ Fully extended hash array
  CREATE W      16 CELLS ALLOT  \ Holds message block
  1 W !                         \ For compile time endian testing

  HEX
CREATE  K[I]  \ SHA-256 round constants array
428A2F98 , 71374491 , B5C0FBCF , E9B5DBA5 , 3956C25B , 59F111F1 , 923F82A4 , AB1C5ED5 ,
D807AA98 , 12835B01 , 243185BE , 550C7DC3 , 72BE5D74 , 80DEB1FE , 9BDC06A7 , C19BF174 ,
E49B69C1 , EFBE4786 , 0FC19DC6 , 240CA1CC , 2DE92C6F , 4A7484AA , 5CB0A9DC , 76F988DA ,
983E5152 , A831C66D , B00327C8 , BF597FC7 , C6E00BF3 , D5A79147 , 06CA6351 , 14292967 ,
27B70A85 , 2E1B2138 , 4D2C6DFC , 53380D13 , 650A7354 , 766A0ABB , 81C2C92E , 92722C85 ,
A2BFE8A1 , A81A664B , C24B8B70 , C76C51A3 , D192E819 , D6990624 , F40E3585 , 106AA070 ,
19A4C116 , 1E376C08 , 2748774C , 34B0BCB5 , 391C0CB3 , 4ED8AA4A , 5B9CCA4F , 682E6FF3 ,
748F82EE , 78A5636F , 84C87814 , 8CC70208 , 90BEFFFA , A4506CEB , BEF9A3F7 , C67178F2 ,


  DECIMAL

MACRO ]L   " ] LITERAL "

  0 VALUE H[H]  \ Pointer to addr of hash value H for each round

MACRO H[G]  " H[H]  [ 1 CELLS ]L  +"  \ Return G adr
MACRO H[F]  " H[H]  [ 2 CELLS ]L  +"  \ Return F adr
MACRO H[E]  " H[H]  [ 3 CELLS ]L  +"  \ Return E adr
MACRO H[D]  " H[H]  [ 4 CELLS ]L  +"  \ Return D adr
MACRO H[C]  " H[H]  [ 5 CELLS ]L  +"  \ Return C adr
MACRO H[A]  " H[H]  [ 7 CELLS ]L  +"  \ Return A adr

MACRO ror\  " DUP >R [ CELLSIZE \ TUCK - ]L LSHIFT R> LITERAL RSHIFT OR "

: SHAinit ( -)  \ Load initial hash values H0 - H7
  [ HEX ] 6A09E667 ( H0)  BB67AE85 ( H1)  3C6EF372 ( H2)  A54FF53A ( H3)
          510E527F ( H4)  9B05688C ( H5)  1F83D9AB ( H6)  5BE0CD19 ( H7)
  [ DECIMAL ]
  SHAsh  7 0 DO  TUCK  !  CELL+  LOOP  !   \ Put initial hash in SHAsh array
  SHAsh  SHAval  8 CELLS  CMOVE            \ Put copy in SHAval array
  SHAsh  TO  H[H]                          \ Init pointer to last hash value
;

: UpDateHash ( -)  \ Update hash values and load arrays with new values
  SHAsh  SHAval  H[H]                                  \ Place array addresses on stack
  8 0 DO  DUP >R   @   SWAP  DUP >R  @  +  DUP         \ Compute updated hash subvalue
          R@  !  OVER  !  CELL+  R>  CELL+  R>  CELL+  \ Store updated hash subvalue
  LOOP  3DROP                                          \ Clear stack when done
  SHAsh  TO  H[H]                                      \ Init pointer to last subvalue
;

\ ( - n )  n = (E AND F) XOR/OR (~E AND G)  --  can use XOR or OR
MACRO Ch   " H[F] 2@  OVER  AND  SWAP  INVERT  H[G]  @  AND  XOR"

\ ( - n )  n = (A AND B) XOR/OR (A AND C) XOR/OR (B AND C)  --  can use XOR or OR
MACRO Maj  " H[C]  DUP >R  CELL+  2@  OVER  AND  SWAP  R@ @  AND XOR  R> 2@ AND  XOR "

\ ( - n )  T1x = Ch(e,f,g) + Sig1(e) + h
: T1x  Ch  H[E] @  DUP >R  ror\ 6  R@  ror\ 11  XOR  R>  ror\ 25  XOR  +  H[H] @  +  ;

\ ( - n )  T2 = Maj(a,b,c) + Sig0(a)
: T2  Maj  H[A] @  DUP >R  ror\ 2  R@  ror\ 13  XOR  R>  ror\ 22  XOR  +  ;

\ ( x - n )  n = ROR7(X)   XOR  ROR18(X)  XOR  RSH3(X)
: sig0  ( x - n)   DUP >R  ror\  7  R@  ror\ 18  XOR  R>   3 RSHIFT  XOR  ;

\ ( x - n )  n = ROR17(X)  XOR  ROR19(X)  XOR  RSH10(X)
: sig1  ( x - n )  DUP >R  ror\ 17  R@  ror\ 19  XOR  R>  10 RSHIFT  XOR  ;

\ Put two copies of original Wi on stack, keep its address
MACRO Wi@  " DUP  @  TUCK"  ( [Wi] - wi [Wi] wi)

\ Create 2 copies of new Wi' from Wi on stack  ( ..Wi -..Wi' Wi')
MACRO Wi  " 15 PICK  15 PICK  sig0  +  7 PICK  +  2 PICK  sig1  +  DUP "

\ Drop 64 Wi cells from stack
MACRO WiDROP  " 8 0 DO  2DROP 2DROP  2DROP 2DROP  LOOP"  ( W0..W63 - )

\ Add round constant Ki to x, then point to next Ki in array
  0 VALUE Ki    \ Pointer to Ki round constants
: Ki+  ( x - y)  Ki  DUP  CELL+  TO  Ki  @  + ;

MACRO subrnd  " T1x +  Ki+  ( T1)  DUP  H[D]  +!  T2  +  H[G] TO H[H]  H[A]  !"

: SHA256  ( Wadr - )  \ Compute 256-bit SHA-256 hash of 512-bit message block
  K[I]  TO  Ki                                \ Init constants array pointer
  16 0  DO  Wi@  subrnd  CELL+  LOOP  DROP    \ ( W0..W15)  ,original block
  48 0  DO  Wi   subrnd         LOOP  WiDROP  \ ( - )
  UpDateHash
;

: setlen  ( -- )  \ Store count into last two cells
  SHAlen @  DUP  [ CELLSIZE 3 - ]L  RSHIFT  [ W 56 CHARS + ]L !  ( hi)
  3 LSHIFT  [ W 60 CHARS + ]L !  ( lo)
;

: bytes>< ( m -- w )  \ Reverse cell bytes: 12345678 <-> 78563412
  [ HEX ]  DUP >R  18 LSHIFT  R@  FF00 AND  8 LSHIFT  OR
  R@  FF0000 AND  8 RSHIFT  OR  R>  18 RSHIFT  OR  [ DECIMAL ]
;

: cellsreverse  ( adr n -- )  \ Reverse bytes of n cells in array
  0 DO  DUP  @  bytes><  OVER !  CELL+  LOOP  DROP
;

W C@ [IF]  \ if little ENDIAN, e.g. PCs
      MACRO endian16 "  DUP  16  cellsreverse "  ( adr -- adr )
      MACRO endian14 "  DUP  14  cellsreverse "  ( adr -- adr )
[ELSE]      \ if big ENDIAN, e.g. Macs
      MACRO endian16 " "  ( adr -- adr )  \ Do nothing
      MACRO endian14 " "  ( adr -- adr )  \ Do nothing
[THEN]

\ Do all 64 byte blocks leaving remainder block
: hashfullblocks ( adr1 count1 --  adr2 count2 )
  DUP  -64  AND  ( count 63 > - ?)       \ Maskoff byte count, if > 63 then
  IF  DUP >R  6 RSHIFT     ( adr #/64 )  \ Compute number of full 64 byte blocks
      0 DO  DUP  endian16  ( adr adr  )  \ Endian adjust block cells if necessary
            SHA256  64 +   ( adr+64   )  \ Hash block, set address to next blcok
      LOOP                 ( adr'     )  \ Do for all full 64 byte blocks
      R> 63 AND            ( adr2 #   )  \ Maskoff partial block bytes count: 0-63
  THEN                     ( adr2 cnt2)  \ Leave address and count for last block
;

: hashfinal ( addr count -- )  \ Hash partial and/or last block
  DUP >R  W  DUP >R  SWAP  CMOVE        \ Move bytes into block W array
  R> R@ +  128  OVER  C!   ( adr     )  \ Put 80h after last message byte
  CHAR+  55 R@ -           ( adr #   )  \ Compute tentative 0 byte FILL count
  R> 55 >                  ( adr # ? )  \ Is partial block byte count > 55 ?
  IF    8 + 0  FILL             ( -- )  \ If yes, FILL rest of block w/zeroes
        W  endian16  SHA256     ( -- )  \ Endian adjust block if required, then hash
        W  56              ( adr 56  )  \ Now setup last block containing bit count
  THEN                     ( adr #   )
  0 FILL  setlen  W  endian14  SHA256   \ Zero FILL last block, set message bit count
;                               ( -- )  \ Endian adjust, except bit count, then hash

\ Compute SHA256 from a counted buffer of text
: SHAbuffer ( addr count -- )
  SHAinit  DUP  SHAlen !  hashfullblocks  hashfinal
;

