HERE
S" sha256.f" INCLUDED
HERE SWAP - .( SHA256 size is ) . CR
S" sha256display.f"  INCLUDED


\ ====================  SHA-256 Test Suite  =====================
  DECIMAL

\ Load W array with data on stack
: WLoad ( d0..d15 -- )  [ W 15 CELLS + ]L  ( d0..d15 W[15] )
  16 0 DO  TUCK  !  CELL-  LOOP  DROP
;

\ -------------------------------------------------------------
\ EXAMPLE 1:
\ Message: ASCII string 'abc'
\ Hash = ba7816bf 8f01cfea 414140de 5dae2223 b00361a3 96177a9c b410ff61 f20015ad

\ Compute and display hash for ASCII string 'abc'
: EX1  S" abc"  SHAbuffer  HASH. 
 CR ." Must be:"
 CR ." BA7816BF 8F01CFEA 414140DE 5DAE2223 B00361A3 96177A9C B410FF61 F20015AD"
 ;

\ -------------------------------------------------------------
\ EXAMPLE 2:
\ Message:'abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq'
\ Hash = 248d6a61 d20638b8 e5c02693 0c3e6039 a33ce459 64ff2167 f6ecedd4 19db06c1

: EX2a  S" abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq" 
;
: EX2  EX2a  SHAbuffer  HASH.  
 CR ." Must be:"
 CR ." 248D6A61 D20638B8 E5C02693 0C3E6039 A33CE459 64FF2167 F6ECEDD4 19DB06C1"
;

\ -------------------------------------------------------------
\ EXAMPLE 3:
\ Message: 1 million copies of 'a' (61h), (8 million bits)
\ Hash = cdc76e5c 9914fb92 81a1c7e2 84d73e67 f1809a48 a497200e 046d39cc c7112cd0

\ Load block of all 'a's (61h), must hash 15,625 times
: EX3a  W  64  [CHAR] a  FILL  ;

\ Last message block: 1st bit a '1', bit-count = 8 million
: EX3b  [ HEX ] 80000000 0 0 0 0 0 0 0 0 0 0 0 0 0 0
        [ DECIMAL ] 8000000 WLoad
;
\ Do hash for message of 1 million copies of ASCII 'a' (61h)
: EX3  SHAinit  EX3a  15625 0 DO W SHA256 LOOP  EX3b  W SHA256 HASH. 
CR ." Must be:"
CR ." CDC76E5C 9914FB92 81A1C7E2 84D73E67 F1809A48 A497200E 046D39CC C7112CD0"
;

\ -------------------------------------------------------------
\ Message: 10,000 SPACES 'BL' (20h), (80,000 bits)
\ Hash = 7c99bc05 0d0856d0 5a0d84b5 44718f5f 2764cf6e fc939d70 42086f27 19fe8ec2

\ Load block of all "BL' (20h), hash 156 full blocks + 16 bytes
: EX4a  W  64  BL  FILL  ;

\ Last message block: 16 SPACES, bit-count = 80,000
: EX4b  [ HEX ] 20202020 20202020 20202020 20202020 80000000
                0 0 0 0 0 0 0 0 0 0 [ DECIMAL ] 80000 WLoad
;

\ Do hash for message of 10,000 SPACES 'BL' (20h)
: EX4  SHAinit  EX4a  156 0 DO  W SHA256  LOOP  EX4b  W SHA256 HASH. 
CR ." Must be:"
CR ." 7C99BC05 0D0856D0 5A0D84B5 44718F5F 2764CF6E FC939D70 42086F27 19FE8EC2"
;

\ -------------------------------------------------------------
\ Message: blank string ''
\ Hash = e3b0c442 98fc1c14 9afbf4c8 996fb924 27ae41e4 649b934c a495991b 7852b855


: SHATest ( -- )
  CR ." SHA-256 test suite:"
  S" "  SHAbuffer  HASH.  S" "  QuoteString
  CR ." Must be: E3B0C442 98FC1C14 9AFBF4C8 996FB924 27AE41E4 649B934C A495991B 7852B855"
  EX1  S" abc"  QuoteString
  EX2  EX2a     QuoteString
  EX3  S" 1 million copies of ASCII 'a' (61h)" TYPE
  EX4  S" 10,000 copies of ASCII BL (20h)" TYPE  CR
  
;

\ ===========  SwiftForth specific performance test ===========

  DECIMAL 100 VALUE N#

: [EX1]  S" abc"  SHAbuffer  ;
: [EX2]  EX2a     SHAbuffer  ;
: [EX3]  SHAinit  EX3a  15625 0 DO W SHA256 LOOP EX3b W SHA256 ;

: ucounter GetTickCount ;
: utimer GetTickCount SWAP - . ;

: test1  [ DECIMAL ]
  CR ." SHA-256 test for EX1 for " N# . ." loops in milliseconds is "
  ucounter  N# 0 DO  [EX1]  LOOP  utimer
;

: test2  [ DECIMAL ]
  CR ." SHA-256 test for EX2 for " N# . ." loops in milliseconds is "
  ucounter  N# 0 DO  [EX2]  LOOP  utimer
;

: test3  [ DECIMAL ]
  CR ." SHA-256 test for EX3 for " N# . ." loops in milliseconds is "
  ucounter  N# 0 DO  [EX3]  LOOP  utimer
;
