\ AVL-tree implementation
\ Author: Nicholas Nemtsev
\ avlNode -    ,     
\     :
\   Compare ( x -- n)       ;  
\                          ,  x < 
\                        ,  x > 
\                        ,   = 
\   toString ( -- a u) -    
\ avlTree -  ,     avl-
\         (   )
\   :
\     Insert ( x -- )           ( )
\     Insert0 ( x --)            
\     Find   ( x -- node/0)    
\     print1                       :
\                             , ,  
\     print2                 ,  ,  

REQUIRE { ~nn/lib/locals.f
REQUIRE .R lib/ext/core-ext.f

REQUIRE CLASS: ~nn/class/class.f

CLASS: avlNode
    var vLeft
    var vRight
    var vHeight

CONSTR: init ( -- )  vLeft 0!  vRight 0!  vHeight 0! ;

VM: Compare ( x -- n ) DROP 0 ;
VM: toString ( -- a u) S" " ;
;CLASS

CLASS: avlTree 
    var vRoot
    var vNodeClass

WITH avlNode        

    : (free) @ ?DUP
        IF DUP => vLeft RECURSE
           DUP => vRight RECURSE 
           DELETE
        THEN ;
    DESTR: free vRoot (free)  ;
    CONSTR: init vNodeClass ! vRoot 0! ;

    : (find) ( obj tree -- tree ?)
        DUP @ ?DUP 0= ( .S) IF NIP FALSE ( .S) EXIT THEN
        >R OVER R@ => Compare ?DUP 0= IF RDROP NIP TRUE EXIT THEN \ found
        ( obj tree ? --) NIP R> SWAP
        0< IF => vLeft ELSE => vRight THEN
        RECURSE 
    ;

    : (end-order) ( xt tree -- )
        ?DUP 0= IF DROP EXIT THEN
        2DUP => vLeft @ RECURSE
        2DUP SWAP EXECUTE
             => vRight @ RECURSE
    ;

    : (pre-order) ( xt tree -- )
        ?DUP 0= IF DROP EXIT THEN
        2DUP SWAP EXECUTE
        2DUP => vLeft @ RECURSE
             => vRight @ RECURSE
    ;

    M: endOrder ( xt --) vRoot @ (end-order) ;
    M: preOrder ( xt --) vRoot @ (pre-order) ;

    M: Find ( x -- node/0 ) vRoot (find) DROP @ ;

    : (find-min) ( tree -- tree1 )
        DUP @ ?DUP 0= ( .S) IF ( .S) EXIT THEN
        DUP => vLeft @ 0= IF DROP EXIT THEN
        NIP => vLeft RECURSE 
    ;

    : (find-max) ( tree -- tree1 )
        DUP @ ?DUP 0= ( .S) IF ( .S) EXIT THEN
        DUP => vRight @ 0= IF DROP EXIT THEN
        NIP => vRight RECURSE 
    ;


    M: FindMin ( -- node )  vRoot (find-min) @ ;
    M: FindMax ( -- node )  vRoot (find-max) @ ;
    
    M: Insert0 ( obj -- )
        DUP vRoot (find) 
        0=
        IF
            !
        ELSE 
            2DROP
        THEN
    ;

    : Height ( P -- n ) ?DUP 0= IF -1 ELSE => vHeight @ THEN ;
    : Left => vLeft @ ;
    : Left! => vLeft ! ;
    : Right => vRight @ ;
    : Right! => vRight ! ;
    : Height! => vHeight ! ;
    : 2Height! { K1 K2 -- }
        K2 Left Height  K2 Right Height MAX 1+ K2 Height!
        K1 Left Height  K2 Height MAX 1+ K1 Height!
    ;

    : SingleRotateWithLeft { K2 \ K1 -- K1 }
        K2 Left TO K1
        K1 Right K2 Left!
        K2 K1 Right! 
        K1 K2 2Height!
        K1
    ;

    : SingleRotateWithRight { K1 \ K2 -- K2 }
        K1 Right TO K2
        K2 Left K1 Right!
        K1 K2 Left!
        K2 K1 2Height!
        K2
    ;

    : DoubleRotateWithLeft { K3 -- node }
        K3 Left SingleRotateWithRight K3 Left!
        K3 SingleRotateWithLeft
    ;

    : DoubleRotateWithRight { K1 -- node }
        K1 Right SingleRotateWithLeft K1 Right!
        K1 SingleRotateWithRight
    ;

    : (insert) { X T -- tree }
        T 0= 
        IF 
           X vNodeClass @ NEW TO T
        ELSE
            X T => Compare DUP 0<
            IF DROP
                X T Left RECURSE T Left!
                T Left Height T Right Height - 2 =
                IF
                    X T Left => Compare 0<
                    IF                   
                        T SingleRotateWithLeft TO T
                    ELSE
                        T DoubleRotateWithLeft TO T
                    THEN
                THEN
            ELSE 
            0 >
            IF
                X T Right RECURSE T Right!
                T Right Height T Left Height - 2 =
                IF
                    X T Right => Compare 0 >
                    IF
                        T SingleRotateWithRight TO T
                    ELSE
                        T DoubleRotateWithRight TO T
                    THEN
                THEN
            THEN THEN 
        THEN
        T Left Height T Right Height MAX 1+ T Height!
        T
    ;

    M: Insert ( x -- ) vRoot @ (insert) vRoot ! ;
    
    M: delete { tree \ node left right moved -- }
        \    (     ).
        tree @ TO node
        node => vLeft @ TO left
        node => vRight @ TO right
        left 0= right 0= OR 
        IF \        
            left ?DUP 0= IF right THEN tree !
        ELSE
            \   .
            \      
            node => vLeft
            BEGIN DUP @ => vRight DUP @ WHILE
                NIP
            REPEAT
            DROP
            DUP @ TO moved \  
            DUP node => vLeft =
            IF \       .
                right moved => vRight ! \     . . .
            ELSE
                \   
                moved => vLeft @ SWAP ! \    . .    
                left moved => vLeft !
                right moved => vRight !
            THEN
            moved tree !  \   
        THEN
        node DELETE
    ;
    

ENDWITH
;CLASS

REQUIRE \EOF ~nn/lib/eof.f
\EOF

CLASS: numNode <SUPER avlNode
    var vNumber
    CONSTR: init ( n -- ) vNumber ! init ;
    VM: Compare ( x -- n ) vNumber @ - ;
    VM: toString ( -- a u) vNumber @ S>D DUP >R <# #S R> SIGN #> ;
;CLASS


REQUIRE RANDOM ~nn/lib/ran4.f
REQUIRE [NONAME ~nn/lib/noname.f
~nn/lib/utils/avlprint.f

numNode avlTree OBJECT: t1
15 CONSTANT tree-size
: test
    tree-size 0 DO tree-size I - t1 Insert LOOP
    t1 print1
    t1 print2
    [NONAME ->CLASS numNode toString TYPE SPACE NONAME] t1 endOrder CR
    [NONAME ->CLASS numNode toString TYPE SPACE NONAME] t1 preOrder CR
    WITH avlNode
    t1 vRoot t1 delete t1 print2
    t1 vRoot @ => vLeft t1 delete t1 print2
    t1 vRoot t1 delete t1 print2
    t1 vRoot t1 delete t1 print2
    t1 vRoot t1 delete t1 print2
    ENDWITH
;

test
