\ File: parameters.f
\ Author: Nicholas Nemtsev mailto:nemtsev@nncron.ru
\ Date: 2003-11-19 19:00
\ Description: read/write parameters set/get controls by parameters
\ Usage: 
\ PARAMETERS: <pars-voc-name>
\   parB <BooleanParameterName> <def-value={ON|OFF}>
\   parN <NumberParameterName> <def-value=number>
\   parS <StringParameterName> <def-value="string">
\   parL <ListParameterName> <def-value="string">
\   parC <ComboParameterName> <def-value="string"> <xt-of-list-maker>
\ ;PARAMETERS
\ ...
\ : load-pars ['] pars-voc-name S" par.ini" parREAD ;
\ : save-pars ['] pars-voc-name S" par.ini" parWRITE ;
\ : set-controls-by-pars ['] pars-voc-name parSETALL ;
\ : set-pars-by-controls ['] pars-voc-name parGETALL ;
\ ... parBooleanParameterName get IF ... THEN
\ ... StringParameterName get INCLUDED
\ ...
\ also see example at end of file

REQUIRE get-string ~nn/lib/getstr.f
REQUIRE EVAL-SUBST ~nn/lib/subst1.f
REQUIRE PLACE lib/ext/string.f
REQUIRE .R lib/ext/core-ext.f
REQUIRE @AZ ~nn/lib/az.f
REQUIRE [NONAME ~nn/lib/noname.f
REQUIRE OPEN/CREATE-FILE ~nn/lib/file.f
REQUIRE CLASS: ~nn/class/class.f
REQUIRE Control ~nn/lib/win/control.f
REQUIRE ComboBox ~nn/lib/win/controls/combobox.f
REQUIRE OFF ~nn/lib/onoff.f
REQUIRE ?FREE2 ~nn/lib/free.f
REQUIRE AppendNode ~nn/lib/list.f
REQUIRE S>NUM ~nn/lib/s2num.f
REQUIRE N>S  ~nn/lib/num2s.f
REQUIRE VOC-FOREACH ~nn/lib/utils/vocs.f

\ type of options ==================================
\ string
\ several strings
\ number
\ boolean

\ actions ==========================================
\ 1. create object with 'ini' prefix
\ 2. create CONSTANT entry in INIOPTIONS vocabulary (ref to object)

VARIABLE parName
VARIABLE parDefVal

: parName! get-string EVAL-SUBST DROP parName ! ;

CLASS: Parameter
    var vPar
    var vDef
    var vIsSet
    var vControl
    CONSTR: init vPar 0! vIsSet OFF ;
    DESTR: free ;
    M: ctrl! vControl ! ;
    M: ctrl@ vControl @ ;
    M: GetName this [ 1 CELLS CFL + CFL + ] LITERAL - >NAME COUNT 3 /STRING ;
    M: WriteName GetName TYPE ;
    M: WriteName: WriteName ." :" SPACE ;
    M: WriteNameBl WriteName SPACE ;
    VM: ReadPar ( read par value from input stream ) ;
    VM: WritePar ( write par name and value to output stream ) ;
    VM: SetPar ( set control by par value ) ;
    VM: GetPar ( set par value by control ) ;
    VM: get ;
    VM: set ;
    M: isSet vIsSet ON ;
    M: isSet? vIsSet @ ;
    M: getPar isSet? IF vPar ELSE vDef THEN ;
;CLASS

CLASS: BoolParameter <SUPER Parameter
    CONSTR: init ( def-value) init vDef ! ;
    VM: get getPar @ ;
    VM: set vPar ! isSet ;
    VM: ReadPar get-boolean set ;
    VM: WritePar WriteNameBl get IF ." ON" ELSE ." OFF" THEN CR ;
    VM: SetPar ctrl@ IF ctrl@  get
        IF ->CLASS CheckBox Checked 
        ELSE ->CLASS CheckBox Unchecked THEN THEN ;
    VM: GetPar ctrl@ IF ctrl@ ->CLASS CheckBox GetCheck set THEN ;
;CLASS

CLASS: NumParameter <SUPER BoolParameter
    VM: ReadPar get-number set ;
    VM: WritePar WriteName: get DECIMAL 1 .R CR ;
    VM: SetPar ctrl@ IF get N>S ctrl@ ->CLASS Edit SetText THEN ;
    VM: GetPar ctrl@ IF ctrl@  ->CLASS Edit GetText S>NUM set THEN ;
;CLASS

CLASS: StrParameter <SUPER Parameter
    CONSTR: init ( a u ) init
        HERE 1+ vDef ! S", 0 C, ;
    VM: get getPar @ ASCIIZ> ;
    VM: set ( a u --) vPar @ ?FREE2 S>ZALLOC vPar ! isSet ;
    VM: ReadPar get-string set ;
    VM: WritePar WriteName: get TYPE CR ;
    VM: SetPar ctrl@ IF get ctrl@ ->CLASS Edit SetText THEN ;
    VM: GetPar ctrl@ IF ctrl@ ->CLASS Edit GetText set THEN ;
;CLASS

CLASS: ListParameter <SUPER Parameter
    CONSTR: init ( a u ) init
        HERE vDef ! 
        0 , HERE CELL+ 1+ ,
        S", 0 C, ;
    VM: get getPar ( list) ;
    VM: set ( a u --) S>ZALLOC vPar AppendNode isSet ;
    VM: ReadPar get-string set ;
    VM: WritePar 
        [NONAME
            WriteName:
            NodeValue ASCIIZ> TYPE CR
        NONAME] get DoList
    ;
    VM: SetPar 
        [NONAME
            NodeValue ASCIIZ>
            vControl @ ->CLASS ListBox Add
        NONAME] get DoList ;
    VM: GetPar
        vPar OFF
        vControl @ ->CLASS ListBox Length
        0 ?DO
              PAD I vControl @ ->CLASS ListBox Get set
        LOOP
    ;
;CLASS

CLASS: ComboParameter <SUPER StrParameter
    var vListMakerXT
    var vListItems
    CONSTR: init ( a u xt ) vListMakerXT ! init ;
    M: addStr ( a u -- ) ctrl@ ->CLASS ComboBox Add ;
    M: getIndex ( a u -- n ) ctrl@ ->CLASS ComboBox GetIndex ;
    M: setCur ctrl@ ->CLASS ComboBox Current! ;
    VM: SetPar { \ list -- }
        ctrl@
        IF
            vListItems 0!
            ctrl@ ->CLASS ComboBox Clear
            vListItems vListMakerXT @ EXECUTE
            [NONAME
                NodeValue ASCIIZ> addStr
            NONAME] vListItems DoList
            get getIndex CB_ERR = 
            IF get addStr THEN
            get getIndex setCur
        THEN
        ;
    VM: GetPar ctrl@ IF ctrl@ ->CLASS ComboBox GetText set THEN ;
;CLASS


VECT PARAMETERS-VOC ' NOOP TO PARAMETERS-VOC
: >PARAMETERS ALSO PARAMETERS-VOC DEFINITIONS ;
: PARAMETERS> PREVIOUS DEFINITIONS ;

: PARAMETERS: 
    VOCABULARY
    LATEST NAME> TO PARAMETERS-VOC ;
: ;PARAMETERS ;

VARIABLE parNeedColon
: parColon parNeedColon @ IF S" :" ELSE S" " THEN ;
: parEntry
    S" >PARAMETERS par%parName @AZ% SELF CONSTANT %parName @AZ%%parColon% PARAMETERS>"
    EVAL-SUBST  EVALUATE
;

: parObject S" OBJECT: par%parName @AZ%" EVAL-SUBST EVALUATE ;

: parA
\ alias
;

: parB ( | name default-value -- )
\ boolean parameter
    parNeedColon OFF
    parName! get-boolean
    BoolParameter parObject
    parEntry
;

: parS ( | name default-value -- )
    parNeedColon ON
    parName! get-string
    StrParameter parObject
    parEntry
;
: parL
    parNeedColon ON
    parName! get-string
    ListParameter parObject
    parEntry
;
: parN
    parNeedColon ON
    parName! get-number
    NumParameter parObject
    parEntry
;
: parC ( | name default-value list-maker -- )
    parNeedColon ON
    parName! get-string ' 
    ComboParameter parObject
    parEntry
;

: parWRITE { a u -- }
\   
    H-STDOUT >R
    a u R/W OPEN/CREATE-FILE ?DUP 0=
    IF
        TO H-STDOUT
        [NONAME NodeValue ->CLASS Parameter WritePar NONAME] VOC-FOREACH
        H-STDOUT FILE-POSITION THROW
        H-STDOUT RESIZE-FILE THROW
        H-STDOUT CLOSE-FILE DROP
    ELSE
        ." Error # " . ."  Can't open " a u TYPE CR
        DROP
    THEN
    R> TO H-STDOUT
;
: parREAD { xt-voc a u \ pre-save -- }
\   
    ['] <PRE> BEHAVIOR TO pre-save
    [NONAME
        NextWord SFIND
        IF 
            EXECUTE ->CLASS Parameter ReadPar
        ELSE 2DROP THEN
        1 PARSE 2DROP 
    NONAME] TO <PRE>
    ONLY xt-voc EXECUTE
    a u ['] INCLUDED CATCH IF 2DROP THEN
    ONLY FORTH DEFINITIONS
    pre-save TO <PRE>
;

: parSETALL ( xt-voc -- ) [NONAME NodeValue ->CLASS Parameter SetPar NONAME] VOC-FOREACH ;
: parGETALL ( xt-voc -- ) [NONAME NodeValue ->CLASS Parameter GetPar NONAME] VOC-FOREACH ;

REQUIRE \EOF ~nn/lib/eof.f
\EOF

\ iniSS Crontab nncron.tab
\ iniN Delay 100

: make-lang-list { list -- }
\  -      res  
    S" Russian" list AddNode
    S" English" list AddNode
    S" German" list AddNode
;

PARAMETERS: backupINI
    parB Reload OFF
    parB RunAtStartup ON
    parL Crontab nncron.tab
    parS LogName log\nncron.log
    parN DelayValue 100
    parN CompleteTaskTimeout 10000
    parN QueryStartTimeout 30
    parC Language Russian make-lang-list
;PARAMETERS

\ iniWRITE
' backupINI S" nn.ini" parREAD
' backupINI S" nn.ini" parWRITE

parReload WritePar
parRunAtStartup WritePar
parCrontab WritePar
parLogName WritePar
parDelayValue WritePar
parCompleteTaskTimeout WritePar
parQueryStartTimeout WritePar
parLanguage WritePar