REQUIRE MODULE: ~nn/lib/spf4/spf_modules.f
REQUIRE Z" ~nn/lib/az.f
REQUIRE GET ~nn/lib/mutex.f
REQUIRE WZ" ~nn/lib/u/util/wz.f
REQUIRE 2VARIABLE lib/ext/double.f
REQUIRE Win2k? ~nn/lib/winver.f
REQUIRE #define ~nn/lib/define.f
REQUIRE CLASS: ~nn/class/class.f
REQUIRE [NONAME ~nn/lib/noname.f
REQUIRE DoList ~nn/lib/list.f
REQUIRE QUEUE ~nn/lib/utils/queue.f
\ WINAPI: lstrcmpiW KERNEL32.DLL

\  HERE 
MODULE: PERF-COUNTERS

REQUIRE /PERF_COUNTER_DEFINITION ~nn/lib/win/sys/perfdef.f

100 1024 * CONSTANT TOTALBYTES
1024 CONSTANT BYTEINCREMENT
#define SYSTEM_OBJECT_INDEX                 2   \   // 'System' object
#define PROCESS_OBJECT_INDEX                230 \   // 'Process' object
#define PROCESSOR_OBJECT_INDEX              238 \       // 'Processor' object
#define TOTAL_PROCESSOR_TIME_COUNTER_INDEX  240 \   // '% Total processor time' counter (valid in WinNT under 'System' object)
#define PROCESSOR_TIME_COUNTER_INDEX        6   \   // '% processor time' counter (for Win2K/XP)

VARIABLE Buffer
VARIABLE BufferSize
: Buffer0! Buffer @ BufferSize @ ERASE ;
: ?BufferAlloc Buffer @ 0= IF TOTALBYTES ALLOCATE THROW Buffer ! TOTALBYTES BufferSize ! THEN ;
: Buffer.Realloc ( newsize -- )
    Buffer @ OVER RESIZE THROW Buffer !
    BufferSize !
;
: QueryPerformanceData { \ newBufferSize -- pPerfData }
\      ." QueryPerformanceData" CR
\    GetTickCount >R
\   void QueryPerformanceData(PERF_DATA_BLOCK **pPerfData)
\   {
\       //
\       // Since i want to use the same allocated area for each query,
\       // i declare CBuffer as static.
\       // The allocated is changed only when RegQueryValueEx return ERROR_MORE_DATA
\       //
    ?BufferAlloc
\       static CBuffer Buffer(TOTALBYTES);
\       DWORD BufferSize = Buffer.GetSize();
\       LONG lRes;
    Buffer0!
\       Buffer.Reset();

    BEGIN
        BufferSize @ TO newBufferSize
        AT newBufferSize Buffer @ 0 0 Z" Global" HKEY_PERFORMANCE_DATA RegQueryValueExA
        ERROR_MORE_DATA =
\        ( HKEY_PERFORMANCE_DATA,"Global",NULL,NULL,Buffer,&BufferSize )) == ERROR_MORE_DATA
\       while( (lRes = RegQueryValueEx( HKEY_PERFORMANCE_DATA,"Global",NULL,NULL,Buffer,&BufferSize )) == ERROR_MORE_DATA )
    WHILE
\        ." realloc" CR
         newBufferSize Buffer.Realloc
\           // Get a buffer that is big enough.
\           BufferSize += BYTEINCREMENT;
\           Buffer.Realloc(BufferSize);
    REPEAT
    Buffer @
\    GetTickCount R> - ." QueryPerformanceData time is " . BufferSize @ . CR
\       *pPerfData = (PPERF_DATA_BLOCK) Buffer.m_pBuffer;
;

: FirstObject ( PerfData -- first)
\   PPERF_OBJECT_TYPE FirstObject( PPERF_DATA_BLOCK PerfData )
\       return( (PPERF_OBJECT_TYPE)((PBYTE)PerfData + PerfData->HeaderLength) );
    DUP pdbHeaderLength @ +
;
: NextObject ( PerfObj -- next)
\   PPERF_OBJECT_TYPE NextObject( PPERF_OBJECT_TYPE PerfObj )
\       return( (PPERF_OBJECT_TYPE)((PBYTE)PerfObj + PerfObj->TotalByteLength) );
    DUP potTotalByteLength @ +
;

: FirstCounter ( PerfObj -- first)
\   PPERF_COUNTER_DEFINITION FirstCounter( PPERF_OBJECT_TYPE PerfObj )
\       return( (PPERF_COUNTER_DEFINITION) ((PBYTE)PerfObj + PerfObj->HeaderLength) );
    DUP potHeaderLength @ +
;

: NextCounter ( PerfCntr -- next)
\   PPERF_COUNTER_DEFINITION NextCounter( PPERF_COUNTER_DEFINITION PerfCntr )
\       return( (PPERF_COUNTER_DEFINITION)((PBYTE)PerfCntr + PerfCntr->ByteLength) );
    DUP pcdByteLength @ +
;

: FirstInstance ( PerfObj -- first)
\   PPERF_INSTANCE_DEFINITION FirstInstance( PPERF_OBJECT_TYPE PerfObj )
\       return( (PPERF_INSTANCE_DEFINITION)((PBYTE)PerfObj + PerfObj->DefinitionLength) );
    DUP potDefinitionLength @ +
;

: NextInstance ( PerfInst -- next)
\   PPERF_INSTANCE_DEFINITION NextInstance( PPERF_INSTANCE_DEFINITION PerfInst )
\       PPERF_COUNTER_BLOCK PerfCntrBlk;
\       PerfCntrBlk = (PPERF_COUNTER_BLOCK)((PBYTE)PerfInst + PerfInst->ByteLength);
    DUP pidByteLength @ +
\       return( (PPERF_INSTANCE_DEFINITION)((PBYTE)PerfCntrBlk + PerfCntrBlk->ByteLength) );
    DUP pcbByteLength @ +
;

: LookForInstance { pPerfObj pPerfInst pInstanceName \ bstrInstance -- pCounterBlock }
\          _bstr_t bstrInstance;
\          _bstr_t bstrInputInstance = pInstanceName;
    pPerfObj potNumInstances @ 0
    ?DO
\          for( int k=0; k < pPerfObj->NumInstances; k++ )
\          {
        pPerfInst pPerfInst pidNameOffset @ + TO bstrInstance
\              bstrInstance = (wchar_t *)((PBYTE)pPerfInst + pPerfInst->NameOffset);
\        bstrInstance WASCIIZ> DUMP CR
        bstrInstance pInstanceName lstrcmpiW 0=
        IF
\              if (!stricmp((LPCTSTR)bstrInstance, (LPCTSTR)bstrInputInstance))
\              {
            pPerfInst pPerfInst pidByteLength @ + 
            UNLOOP
            EXIT
\                  pCounterBlock = (PPERF_COUNTER_BLOCK) ((LPBYTE) pPerfInst + pPerfInst->ByteLength);
\                  break;
\              }
        THEN
\              // Get the next instance.
        pPerfInst NextInstance TO pPerfInst 
\              pPerfInst = NextInstance( pPerfInst );
\          }
    LOOP
    0
;

: GetCounterValue3 { pPerfObj dwCounterIndex pInstanceName \ pPerfCntr pPerfInst pCounterBlock -- d }
\   T GetCounterValue(PPERF_OBJECT_TYPE pPerfObj, DWORD dwCounterIndex, LPCTSTR pInstanceName)
\       PPERF_COUNTER_DEFINITION pPerfCntr = NULL;
\       PPERF_INSTANCE_DEFINITION pPerfInst = NULL;
\       PPERF_COUNTER_BLOCK pCounterBlock = NULL;
        0 TO pPerfInst
        0 TO pCounterBlock

\       // Get the first counter.
    
    pPerfObj FirstCounter TO pPerfCntr
    
\       pPerfCntr = FirstCounter( pPerfObj );

\       // Look for the index of '% Total processor time'

    pPerfObj potNumCounters @ 0
    ?DO
\       for( DWORD j=0; j < pPerfObj->NumCounters; j++ )
        pPerfCntr pcdCounterNameTitleIndex @ dwCounterIndex =
        IF
\           if (pPerfCntr->CounterNameTitleIndex == dwCounterIndex)
\               break;
            LEAVE
        ELSE

            pPerfCntr NextCounter TO pPerfCntr 
        THEN
\           // Get the next counter.
        
\           pPerfCntr = NextCounter( pPerfCntr );
    LOOP
    pPerfObj potNumInstances @ PERF_NO_INSTANCES =
    IF
\       if( pPerfObj->NumInstances == PERF_NO_INSTANCES )       
\       {
            pPerfObj pPerfObj potDefinitionLength @ + TO pCounterBlock
\           pCounterBlock = (PPERF_COUNTER_BLOCK) ((LPBYTE) pPerfObj + pPerfObj->DefinitionLength);
\       }
    ELSE
\       else
\       {
            pPerfObj FirstInstance TO pPerfInst 
\           pPerfInst = FirstInstance( pPerfObj );
        
\           // Look for instance pInstanceName
            pPerfObj pPerfInst pInstanceName LookForInstance TO pCounterBlock
\       }
    THEN
    pCounterBlock
    IF
\       if (pCounterBlock)
\       {
\           T *lnValue = NULL;
        pCounterBlock pPerfCntr pcdCounterOffset @ +
        2@ SWAP
\           lnValue = (T*)((LPBYTE) pCounterBlock + pPerfCntr->CounterOffset);
\           return *lnValue;
\       }
    ELSE
        -1.
    THEN
\       return -1;
;

VARIABLE isPerfData
: GetCounterValue
{ dwObjectIndex dwCounterIndex pInstanceName \ pPerfData pPerfObj -- d pPerfData }
\   T GetCounterValue(PERF_DATA_BLOCK **pPerfData, DWORD dwObjectIndex, DWORD dwCounterIndex, LPCTSTR pInstanceName = NULL)
    isPerfData @ ?DUP 0= IF QueryPerformanceData DUP isPerfData ! THEN TO pPerfData
\       QueryPerformanceData(pPerfData);
    0 TO pPerfObj
\       PPERF_OBJECT_TYPE pPerfObj = NULL;

\       // Get the first object type.
    pPerfData FirstObject TO pPerfObj
\       pPerfObj = FirstObject( *pPerfData );

\       // Look for the given object index
    pPerfData pdbNumObjectTypes @ 0
    ?DO
\       for( DWORD i=0; i < (*pPerfData)->NumObjectTypes; i++ )
        pPerfObj potObjectNameTitleIndex @ dwObjectIndex =
        IF
\           if (pPerfObj->ObjectNameTitleIndex == dwObjectIndex)
                pPerfObj dwCounterIndex pInstanceName GetCounterValue3
                pPerfData
                UNLOOP EXIT
\               lnValue = GetCounterValue(pPerfObj, dwCounterIndex, pInstanceName);
\               break;
        THEN
        pPerfObj NextObject TO pPerfObj
\           pPerfObj = NextObject( pPerfObj );
    LOOP
    0. pPerfData
;

QUEUE POINTER perfQ
VARIABLE ThrId
VARIABLE evStop
VARIABLE RefreshInterval 1000 RefreshInterval !
VARIABLE CPUUsage
VARIABLE CountersList

EXPORT
CLASS: PerfObject
    var vObjectIndex
    var vCounterIndex
    var vInstanceName   \ unicode
    var vPerfData
    var vFirstTime
    dvar vlnNewValue
    dvar vlnOldValue
    dvar vNewPerfTime100nSec
    dvar vOldPerfTime100nSec
    dvar vDeltaPerfTime100nSec
    dvar vlnValueDelta
    var vNeedDelete
    CONSTR: init TRUE vFirstTime ! ;
    M: setInitial ( dlnNewValue perfData -- )
        vPerfData !
        vlnNewValue 2!
        vPerfData @ pdbPerfTime100nSec 2@ vNewPerfTime100nSec 2!
        vFirstTime @
        IF
            FALSE vFirstTime !
            vlnNewValue 2@ vlnOldValue 2!
            vNewPerfTime100nSec 2@ vOldPerfTime100nSec 2!
        ELSE
            vlnNewValue 2@ vlnOldValue 2@ D- vlnValueDelta 2!
            vNewPerfTime100nSec 2@ vOldPerfTime100nSec 2@ D- vDeltaPerfTime100nSec 2!
            vlnNewValue 2@ vlnOldValue 2!
            vNewPerfTime100nSec 2@ vOldPerfTime100nSec 2!
        THEN
    ;
    VM: get ;
    VM: set ;
    VM: ?alert ;
    M: Add perfQ SELF 0<> IF this perfQ Put THEN ;
    M: Del vNeedDelete ON perfQ SELF 0<> IF this perfQ Put THEN ;
;CLASS

CLASS: CPUUsageObject <SUPER PerfObject
    var vCPUUsage
    CONSTR: init
        init
        Win2k? 
        IF 
            WZ" _Total" vInstanceName !
            PROCESSOR_OBJECT_INDEX vObjectIndex !
            PROCESSOR_TIME_COUNTER_INDEX vCounterIndex !
        ELSE 
            WZ" "  vInstanceName !
            SYSTEM_OBJECT_INDEX vObjectIndex !
            TOTAL_PROCESSOR_TIME_COUNTER_INDEX vCounterIndex !
        THEN 
    ;
    VM: get vCPUUsage @ ;
    VM: set
        vDeltaPerfTime100nSec 2@ OR 
        IF
            100 vlnValueDelta 2@ D>S 100 vDeltaPerfTime100nSec 2@ D>S */ -
        ELSE 0 THEN
        DUP 0<
        IF DROP 0 THEN
        vCPUUsage !
\        ." vCPUUsage = " vCPUUsage @ . CR
    ;
;CLASS

CLASS: ProcCPUUsageObject <SUPER CPUUsageObject
    CONSTR: init ( a u -- )
        init
        S>UNICODE DROP vInstanceName !
        PROCESS_OBJECT_INDEX vObjectIndex !
        PROCESSOR_TIME_COUNTER_INDEX vCounterIndex !
    ;
    VM: set
        vDeltaPerfTime100nSec 2@ OR 
        IF
            vlnValueDelta 2@ D>S 100 vDeltaPerfTime100nSec 2@ D>S */
        ELSE 0 THEN
        DUP 0<
        IF DROP 0 THEN
        vCPUUsage !
\        ." vCPUUsage = " vCPUUsage @ . CR
    ;

;CLASS

PerfObject POINTER pPerfObject

: ObjectIndex Win2k? IF PROCESSOR_OBJECT_INDEX ELSE SYSTEM_OBJECT_INDEX THEN ;
: CpuUsageIndex Win2k? IF PROCESSOR_TIME_COUNTER_INDEX ELSE TOTAL_PROCESSOR_TIME_COUNTER_INDEX THEN ;

: PerfRenewCounters
    isPerfData 0!
    [NONAME
        NodeValue ( ." PerfRenewCounters: " DUP . CR) TO pPerfObject
        pPerfObject vObjectIndex @ 
        pPerfObject vCounterIndex @ 
        pPerfObject vInstanceName @
            GetCounterValue ( perfData dnew -- ) pPerfObject setInitial
        pPerfObject set
        pPerfObject ?alert
    NONAME]
    CountersList DoList
;

: SetCPUUsageOfProcs
;
CPUUsageObject OBJECT: TotalCPUUsage

: ?Add/DelPerfCounter ( -- )
    perfQ CanGet?
    IF 
\        ." AddToPerfCountersList before" CR
        perfQ Get ( DUP . CR) 
        DUP ->CLASS PerfObject vNeedDelete @
        IF
            DUP CountersList DelNode
            GLOBAL DELETE LOCAL
        ELSE
            CountersList AddNode 
        THEN
\        ." AddToPerfCountersList after"  CR
    THEN
;

:NONAME
    16 QUEUE NEW TO perfQ
    CountersList 0!

    TotalCPUUsage Add
    
    GetCurrentThreadId ThrId !
    0 0 1 0 CreateEventA evStop !
    Buffer 0!
    BEGIN
        ?Add/DelPerfCounter
        
        PerfRenewCounters
        TotalCPUUsage get CPUUsage !
        
        \ RefreshInterval @ PAUSE
        RefreshInterval @ evStop @ WAIT OR 
        IF  HKEY_PERFORMANCE_DATA RegCloseKey DROP 
            evStop @ CLOSE-FILE DROP
            ThrId 0!
            Buffer @ ?DUP IF FREE DROP THEN
            0 TO perfQ
            EXIT 
        THEN
    AGAIN
; TASK: PERF-MONITOR-TASK

: ?START-PERF-MONITOR
    ThrId @ 0=
    IF 0 PERF-MONITOR-TASK START CLOSE-FILE DROP 100 PAUSE THEN
;

\ EXPORT
: ?STOP-PERF-MONITOR
    ThrId @
    IF
        evStop @ SetEvent DROP
        1000 PAUSE
    THEN
;

: NEW-PERF-OBJECT ( class -- object ) GLOBAL NEW LOCAL DUP ->CLASS PerfObject Add ;
: GetCpuUsage ( -- n )
    WinNT? 0= IF -1 EXIT THEN
    
    ?START-PERF-MONITOR

    CPUUsage @
;    

;MODULE

\  .( Size of perf is ) HERE SWAP - . CR 

REQUIRE \EOF ~nn/lib/eof.f
\EOF

: x ?STOP-PERF-MONITOR ;

CLASS: AlertOnExceeding <SUPER ProcCPUUsageObject 
    var vUpper
    CONSTR: init ( n a u -- )
        init
        vUpper !
    ;
    VM: ?alert get vUpper @ > IF ." AlertOnExceeding"  CR THEN  ;
;CLASS

: test { \ tm spf -- }
    ?START-PERF-MONITOR
    S" taskmgr" ProcCPUUsageObject NEW-PERF-OBJECT TO tm
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
\      S" taskmgr" ProcCPUUsageObject NEW  AddToPerfContersList
    
    16 S" spf375" AlertOnExceeding NEW-PERF-OBJECT TO spf
    
    BEGIN
      ." CPU Usage: " GetCpuUsage . ." % " 
      ." taskmgr: " tm ->CLASS ProcCPUUsageObject get . 
      ." spf: " spf ->CLASS ProcCPUUsageObject get . 
      13 EMIT
        1000 PAUSE
        KEY? IF ?STOP-PERF-MONITOR KEY 27 = IF BYE THEN  EXIT THEN
    AGAIN
;
test